"""
Copyright 2021 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from os import environ
import json
import string
import numpy as np
import sys
import pandas as pd
import tensorflow as tf
import datetime
from tensorflow.python import keras
from keras.datasets import imdb
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import LSTM
from keras.layers.embeddings import Embedding
from keras.preprocessing import sequence
from keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.models import load_model
from nltk import word_tokenize
from tensorflow.python.keras import Model

import time

current_dir = os.path.dirname(__file__)
model_filename = os.path.join(current_dir, "data", "LSTM.h5")
data_filename = os.path.join(current_dir, "data", "imdb.npz")
word_dict_path = os.path.join(current_dir, "data", "imdb_word_index.json")
predict_file = os.path.join(current_dir, "data", "IMDB.csv")

# set the seed
np.random.seed(7)
# load the dataset top n words only
top_words = 5000
max_review_length = 500

def preprocessing(filename, maxlen, top_words, word_dict_path=None):
    noisy_char = '!\"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n'
    table = str.maketrans(noisy_char, ' '*len(noisy_char))
    data = pd.read_csv(filename, sep=',')
    data.head()
    word_dict = json.load(open(word_dict_path, 'r'))
    X = []
    for review in data['review'].values:
        review = review.translate(table).replace("  ", " ")
        seq = []
        for word in word_tokenize(review):
            if word.lower() not in word_dict:
                continue
            index = word_dict[word.lower()] + 3
            if index >= top_words:
                continue
            seq.append(index)
        X.append(seq)

    X = pad_sequences(X, maxlen=maxlen)
    tf.logging.info("input shape: {}".format(X.shape))
    return X

print ("Load model")
model = load_model(model_filename, compile = False)
print ("Model ready")
model.summary()

pre_begin = datetime.datetime.now()
is_train = False
# writing the train model and getting input data
if environ.get('RESULT_DIR') is not None:
    output_model_folder = os.path.join(os.environ["RESULT_DIR"], "model")
    output_model_path = os.path.join(output_model_folder, model_filename)
else:
    output_model_folder = "model_bak"
    output_model_path = os.path.join("model", model_filename)

os.makedirs(output_model_folder, exist_ok=True)

if environ.get('DATA_DIR') is not None:
    data_folder = os.environ["DATA_DIR"]
    data_path = os.path.join(data_folder, data_filename)
else:
    data_folder = os.getcwd()
    data_path = os.path.join(data_folder, data_filename)
# set the seed
np.random.seed(7)
# load the dataset top n words only
top_words = 5000
(_, _), (X_test, y_test) = imdb.load_data(path=data_path, num_words=top_words)

# truncate and pad input sequences
max_review_length = 500
X_test = sequence.pad_sequences(X_test, maxlen=max_review_length)
np.random.seed(7)
# load the dataset top n words only
pre_end = datetime.datetime.now()

a = np.array(y_test)
a = a.reshape([a.shape[0],1])
print("running on tensorflow cpu")

t1 = time.time()

embedding_vecor_length = 32
embd_model = Model(inputs=model.input, outputs=model.get_layer("embedding").output)

ebd_begin = datetime.datetime.now()
embd_output = embd_model.predict(X_test, batch_size = 1)
ebd_end = datetime.datetime.now()

original_lstm = Model(inputs=model.input, outputs=model.get_layer('lstm').output)
cpu_lstm_begin = datetime.datetime.now()
original_lstm_output = original_lstm.predict(X_test, batch_size = 1024)
cpu_lstm_end = datetime.datetime.now()

lstm_model2 = Sequential()
lstm_model2.add(Dense(1, activation='sigmoid'))
lstm_model2.build((1, 100))
layer_dict_fix = dict([(layer.name, layer) for layer in model.layers])
lstm_model2.layers[0].set_weights(layer_dict_fix['dense'].get_weights())

dense_begin = datetime.datetime.now()
lstm_output_dense = lstm_model2.predict(original_lstm_output, batch_size = 8)
dense_end = datetime.datetime.now()

t2 = time.time()
print("preprocessing time = {:.2f} s".format((pre_end - pre_begin).total_seconds()))
ebd_time = (ebd_end - ebd_begin).total_seconds()
print("embedding time = {:.2f} s".format(ebd_time))
cpu_lstm_time = (cpu_lstm_end - cpu_lstm_begin).total_seconds() - (ebd_end - ebd_begin).total_seconds()
print("cpu lstm time = {:.2f} s".format(cpu_lstm_time))
dense_time = (dense_end - dense_begin).total_seconds()
print("dense lstm time = {:.2f} s".format(dense_time))
print("total cpu time = {:.2f} s".format(ebd_time + cpu_lstm_time + dense_time))
print("total cpu2 time = {:.2f} s".format(t2-t1))

print("Accuracy = {:.2f} %".format(100 * np.sum(np.where(lstm_output_dense > 0.5, 1, 0) == a)/a.size))



# Copyright 2021 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dpuMcParserBase import *


class xvdpu_isa1InstParser(dpuMcParserBase):
    def __init__(self):
        name = "DPUCVDX8G_ISA1"
        super().__init__(name)
        local_dir = os.path.dirname(__file__)
        try:
            self.mc_prser_cdll = CDLL("%s/xvdpu_isa1.so" % local_dir)
            self.process_cpp = self.mc_prser_cdll.process
        except:
            self.mc_prser_cdll = None
            self.process_cpp = None

    def process(self, mc, _debug=False):
        if self.process_cpp != None:
            debug_c = c_bool(_debug)
            mc_c = c_char_p(mc)
            mc_len_c = c_uint(len(mc))

            load_img_size_c = c_uint(0)
            load_para_size_c = c_uint(0)
            save_size_c = c_uint(0)

            # void process(uint8_t *mc, bool debug, uint32_t *load_img_size, uint32_t *load_para_size, uint32_t *save_size)
            self.process_cpp(mc_c, mc_len_c, debug_c, pointer(
                load_img_size_c), pointer(load_para_size_c), pointer(save_size_c))

            self.data["load_img_size"] = load_img_size_c.value
            self.data["load_para_size"] = load_para_size_c.value
            self.data["save_size"] = save_size_c.value
        else:
            pass


register(xvdpu_isa1InstParser())


class xvdpu_isa2InstParser(dpuMcParserBase):
    def __init__(self):
        name = "DPUCVDX8G_ISA2"
        super().__init__(name)
        local_dir = os.path.dirname(__file__)
        try:
            self.mc_prser_cdll = CDLL("%s/xvdpu_isa2.so" % local_dir)
            self.process_cpp = self.mc_prser_cdll.process
        except:
            self.mc_prser_cdll = None
            self.process_cpp = None

    def process(self, mc, _debug=False):
        if self.process_cpp != None:
            debug_c = c_bool(_debug)
            mc_c = c_char_p(mc)
            mc_len_c = c_uint(len(mc))

            load_img_size_c = c_uint(0)
            load_para_size_c = c_uint(0)
            save_size_c = c_uint(0)

            # void process(uint8_t *mc, bool debug, uint32_t *load_img_size, uint32_t *load_para_size, uint32_t *save_size)
            self.process_cpp(mc_c, mc_len_c, debug_c, pointer(
                load_img_size_c), pointer(load_para_size_c), pointer(save_size_c))

            self.data["load_img_size"] = load_img_size_c.value
            self.data["load_para_size"] = load_para_size_c.value
            self.data["save_size"] = save_size_c.value
        else:
            pass


register(xvdpu_isa2InstParser())


class xvdpu_isa3InstParser(dpuMcParserBase):
    def __init__(self):
        name = "DPUCVDX8G_ISA3"
        super().__init__(name)
        local_dir = os.path.dirname(__file__)
        try:
            self.mc_prser_cdll = CDLL("%s/xvdpu_isa3.so" % local_dir)
            self.process_cpp = self.mc_prser_cdll.process
        except:
            self.mc_prser_cdll = None
            self.process_cpp = None

    def process(self, mc, _debug=False):
        if self.process_cpp != None:
            debug_c = c_bool(_debug)
            mc_c = c_char_p(mc)
            mc_len_c = c_uint(len(mc))

            load_img_size_c = c_uint(0)
            load_para_size_c = c_uint(0)
            save_size_c = c_uint(0)

            # void process(uint8_t *mc, bool debug, uint32_t *load_img_size, uint32_t *load_para_size, uint32_t *save_size)
            self.process_cpp(mc_c, mc_len_c, debug_c, pointer(
                load_img_size_c), pointer(load_para_size_c), pointer(save_size_c))

            self.data["load_img_size"] = load_img_size_c.value
            self.data["load_para_size"] = load_para_size_c.value
            self.data["save_size"] = save_size_c.value
        else:
            pass


register(xvdpu_isa3InstParser())




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import platform
import logging
from vaitraceOptions import merge
import vaitraceDefaults


def getx86CpuTscKhz():
    try:
        cmd = 'cat /proc/cpuinfo | grep "cpu MHz"'
        khz = int(float(os.popen(cmd).readlines()[0].split()[3])) * 1000
    except:
        assert()
    return khz


"""
+--------------+-------------------+------------+
| clock_source | x86               | arm        |
+--------------+-------------------+------------+
| VA           | global + XRT      | boot + XRT |
| XAT          | x86-tsc + x86-tsc | boot + XRT |
+--------------+-------------------+------------+
"""


def selectTraceClock(option):
    try:
        clocks = open('/sys/kernel/debug/tracing/trace_clock',
                      'rt').read().strip().split()
    except:
        logging.warning(
            "No such file or directory: '/sys/kernel/debug/tracing/trace_clock'")
        logging.warning("CPU function tracing feature is disabled")
        return "boot"

    va_enabled = option['cmdline_args']['va']
    if va_enabled:
        """VTF do not accept x86-tsc"""
        for c in clocks:
            if c.find("x86-tsc") > 0:
                clocks.remove(c)
                break

    """Ranked by priority"""
    #preferClocks = ["boot", "x86-tsc", "global"]
    preferClocks = ["boot", "global"]
    traceClock = "global"

    for pC in preferClocks:
        for c in clocks:
            if c.find(pC) >= 0:
                traceClock = pC
                break
        else:
            continue
        break

    logging.debug("Use %s as trace clock" % traceClock)

    return traceClock


def getPlatform():
    kRelease = platform.uname().release.split('.')
    machine = platform.uname().machine
    return {'release': [kRelease[0], kRelease[1]], 'machine': machine}


def checkPlatform(plat, option):
    xat_enabled = option['cmdline_args']['xat']
    if xat_enabled and plat.get('machine') == 'x86_64':
        logging.error(".xat format not available for cloud platforms")
        exit(-1)
    return plat


def checkEnv():
    cmds = ["nm --version", "ldd --version", "objdump --version"]

    def checkCmd(cmd):
        try:
            ret = os.system("%s > /dev/null" % cmd)
            if ret != 0:
                return False
        except:
            return False

        return True

    for cmd in cmds:
        if checkCmd(cmd) == False:
            logging.warning("[%s] not exists, function tracer disabled" % (cmd))
            return False

    return True


def checkFtrace():
    return os.path.exists("/sys/kernel/debug/tracing/uprobe_events")


def setting(option: dict):
    if (checkEnv() != True) or (checkFtrace() != True):
        """ Disable function tracer """
        disableFunTracer = {'tracer': {'function': {'disable': True}}}
        merge(option, disableFunTracer)


    traceClock = selectTraceClock(option)
    x86_tsc_khz = 0
    if traceClock == 'x86-tsc':
        x86_tsc_khz = getx86CpuTscKhz()
    plat = checkPlatform(getPlatform(), option)

    runmode = option.get("runmode", vaitraceDefaults.default_runmode)
    if runmode.lower() == "debug":
        os.environ.setdefault("XLNX_ENABLE_DEBUG_MODE", "1")
    else:
        runmode = "normal"
    logging.info("VART will run xmodel in [%s] mode" % runmode.upper())

    globalSetting = {'control': {'traceClock': traceClock, 'x86_tsc_khz': x86_tsc_khz,
                                 'platform': plat, 'runmode': runmode}}
    merge(option, globalSetting)



#Copyright 2021 Xilinx Inc.
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.


#                                 Apache License
#                           Version 2.0, January 2004
#                        https://www.apache.org/licenses/
#
#   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
#
#   1. Definitions.
#      "License" shall mean the terms and conditions for use, reproduction,
#      and distribution as defined by Sections 1 through 9 of this document.
#      "Licensor" shall mean the copyright owner or entity authorized by
#      the copyright owner that is granting the License.
#      "Legal Entity" shall mean the union of the acting entity and all
#      other entities that control, are controlled by, or are under common
#      control with that entity. For the purposes of this definition,
#      "control" means (i) the power, direct or indirect, to cause the
#      direction or management of such entity, whether by contract or
#      otherwise, or (ii) ownership of fifty percent (50%) or more of the
#      outstanding shares, or (iii) beneficial ownership of such entity.
#      "You" (or "Your") shall mean an individual or Legal Entity
#      exercising permissions granted by this License.
#      "Source" form shall mean the preferred form for making modifications,
#      including but not limited to software source code, documentation
#      source, and configuration files.
#      "Object" form shall mean any form resulting from mechanical
#      transformation or translation of a Source form, including but
#      not limited to compiled object code, generated documentation,
#      and conversions to other media types.
#      "Work" shall mean the work of authorship, whether in Source or
#      Object form, made available under the License, as indicated by a
#      copyright notice that is included in or attached to the work
#      (an example is provided in the Appendix below).
#      "Derivative Works" shall mean any work, whether in Source or Object
#      form, that is based on (or derived from) the Work and for which the
#      editorial revisions, annotations, elaborations, or other modifications
#      represent, as a whole, an original work of authorship. For the purposes
#      of this License, Derivative Works shall not include works that remain
#      separable from, or merely link (or bind by name) to the interfaces of,
#      the Work and Derivative Works thereof.
#      "Contribution" shall mean any work of authorship, including
#      the original version of the Work and any modifications or additions
#      to that Work or Derivative Works thereof, that is intentionally
#      submitted to Licensor for inclusion in the Work by the copyright owner
#      or by an individual or Legal Entity authorized to submit on behalf of
#      the copyright owner. For the purposes of this definition, "submitted"
#      means any form of electronic, verbal, or written communication sent
#      to the Licensor or its representatives, including but not limited to
#      communication on electronic mailing lists, source code control systems,
#      and issue tracking systems that are managed by, or on behalf of, the
#      Licensor for the purpose of discussing and improving the Work, but
#      excluding communication that is conspicuously marked or otherwise
#      designated in writing by the copyright owner as "Not a Contribution."
#      "Contributor" shall mean Licensor and any individual or Legal Entity
#      on behalf of whom a Contribution has been received by Licensor and
#      subsequently incorporated within the Work.
#   2. Grant of Copyright License. Subject to the terms and conditions of
#      this License, each Contributor hereby grants to You a perpetual,
#      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#      copyright license to reproduce, prepare Derivative Works of,
#      publicly display, publicly perform, sublicense, and distribute the
#      Work and such Derivative Works in Source or Object form.
#   3. Grant of Patent License. Subject to the terms and conditions of
#      this License, each Contributor hereby grants to You a perpetual,
#      worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#      (except as stated in this section) patent license to make, have made,
#      use, offer to sell, sell, import, and otherwise transfer the Work,
#      where such license applies only to those patent claims licensable
#      by such Contributor that are necessarily infringed by their
#      Contribution(s) alone or by combination of their Contribution(s)
#      with the Work to which such Contribution(s) was submitted. If You
#      institute patent litigation against any entity (including a
#      cross-claim or counterclaim in a lawsuit) alleging that the Work
#      or a Contribution incorporated within the Work constitutes direct
#      or contributory patent infringement, then any patent licenses
#      granted to You under this License for that Work shall terminate
#      as of the date such litigation is filed.
#   4. Redistribution. You may reproduce and distribute copies of the
#      Work or Derivative Works thereof in any medium, with or without
#      modifications, and in Source or Object form, provided that You
#      meet the following conditions:
#      (a) You must give any other recipients of the Work or
#          Derivative Works a copy of this License; and
#      (b) You must cause any modified files to carry prominent notices
#          stating that You changed the files; and
#      (c) You must retain, in the Source form of any Derivative Works
#          that You distribute, all copyright, patent, trademark, and
#          attribution notices from the Source form of the Work,
#          excluding those notices that do not pertain to any part of
#          the Derivative Works; and
#      (d) If the Work includes a "NOTICE" text file as part of its
#          distribution, then any Derivative Works that You distribute must
#          include a readable copy of the attribution notices contained
#          within such NOTICE file, excluding those notices that do not
#          pertain to any part of the Derivative Works, in at least one
#          of the following places: within a NOTICE text file distributed
#          as part of the Derivative Works; within the Source form or
#          documentation, if provided along with the Derivative Works; or,
#          within a display generated by the Derivative Works, if and
#          wherever such third-party notices normally appear. The contents
#          of the NOTICE file are for informational purposes only and
#          do not modify the License. You may add Your own attribution
#          notices within Derivative Works that You distribute, alongside
#          or as an addendum to the NOTICE text from the Work, provided
#          that such additional attribution notices cannot be construed
#          as modifying the License.
#      You may add Your own copyright statement to Your modifications and
#      may provide additional or different license terms and conditions
#      for use, reproduction, or distribution of Your modifications, or
#      for any such Derivative Works as a whole, provided Your use,
#      reproduction, and distribution of the Work otherwise complies with
#      the conditions stated in this License.
#   5. Submission of Contributions. Unless You explicitly state otherwise,
#      any Contribution intentionally submitted for inclusion in the Work
#      by You to the Licensor shall be under the terms and conditions of
#      this License, without any additional terms or conditions.
#      Notwithstanding the above, nothing herein shall supersede or modify
#      the terms of any separate license agreement you may have executed
#      with Licensor regarding such Contributions.
#   6. Trademarks. This License does not grant permission to use the trade
#      names, trademarks, service marks, or product names of the Licensor,
#      except as required for reasonable and customary use in describing the
#      origin of the Work and reproducing the content of the NOTICE file.
#   7. Disclaimer of Warranty. Unless required by applicable law or
#      agreed to in writing, Licensor provides the Work (and each
#      Contributor provides its Contributions) on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#      implied, including, without limitation, any warranties or conditions
#      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#      PARTICULAR PURPOSE. You are solely responsible for determining the
#      appropriateness of using or redistributing the Work and assume any
#      risks associated with Your exercise of permissions under this License.
#   8. Limitation of Liability. In no event and under no legal theory,
#      whether in tort (including negligence), contract, or otherwise,
#      unless required by applicable law (such as deliberate and grossly
#      negligent acts) or agreed to in writing, shall any Contributor be
#      liable to You for damages, including any direct, indirect, special,
#      incidental, or consequential damages of any character arising as a
#      result of this License or out of the use or inability to use the
#      Work (including but not limited to damages for loss of goodwill,
#      work stoppage, computer failure or malfunction, or any and all
#      other commercial damages or losses), even if such Contributor
#      has been advised of the possibility of such damages.
#   9. Accepting Warranty or Additional Liability. While redistributing
#      the Work or Derivative Works thereof, You may choose to offer,
#      and charge a fee for, acceptance of support, warranty, indemnity,
#      or other liability obligations and/or rights consistent with this
#      License. However, in accepting such obligations, You may act only
#      on Your own behalf and on Your sole responsibility, not on behalf
#      of any other Contributor, and only if You agree to indemnify,
#      defend, and hold each Contributor harmless for any liability
#      incurred by, or claims asserted against, such Contributor by reason
#      of your accepting any such warranty or additional liability.
#
#   END OF TERMS AND CONDITIONS
#
#   APPENDIX: How to apply the Apache License to your work.
#
#      To apply the Apache License to your work, attach the following
#      boilerplate notice, with the fields enclosed by brackets "{}"
#      replaced with your own identifying information. (Don't include
#      the brackets!)  The text should be enclosed in the appropriate
#      comment syntax for the file format. We also recommend that a
#      file or class name and description of purpose be included on the
#      same "printed page" as the copyright notice for easier
#      identification within third-party archives.
#
#   Copyright {yyyy} {name of copyright owner}
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
A stacked LSTM with LSTM layers which alternate between going forwards over
the sequence and going backwards.
"""

from typing import Optional, Tuple, Union, List
import torch
from torch.nn.utils.rnn import PackedSequence, pad_packed_sequence, pack_padded_sequence
from allennlp.modules.augmented_lstm import AugmentedLstm
from allennlp.common.checks import ConfigurationError
import time

TensorPair = Tuple[torch.Tensor, torch.Tensor]


class StackedAlternatingLstm(torch.nn.Module):
    """
    A stacked LSTM with LSTM layers which alternate between going forwards over
    the sequence and going backwards. This implementation is based on the
    description in `Deep Semantic Role Labelling - What works and what's next
    <https://homes.cs.washington.edu/~luheng/files/acl2017_hllz.pdf>`_ .

    Parameters
    ----------
    input_size : int, required
        The dimension of the inputs to the LSTM.
    hidden_size : int, required
        The dimension of the outputs of the LSTM.
    num_layers : int, required
        The number of stacked LSTMs to use.
    recurrent_dropout_probability: float, optional (default = 0.0)
        The dropout probability to be used in a dropout scheme as stated in
        `A Theoretically Grounded Application of Dropout in Recurrent Neural Networks
        <https://arxiv.org/abs/1512.05287>`_ .
    use_input_projection_bias : bool, optional (default = True)
        Whether or not to use a bias on the input projection layer. This is mainly here
        for backwards compatibility reasons and will be removed (and set to False)
        in future releases.

    Returns
    -------
    output_accumulator : PackedSequence
        The outputs of the interleaved LSTMs per timestep. A tensor of shape
        (batch_size, max_timesteps, hidden_size) where for a given batch
        element, all outputs past the sequence length for that batch are
        zero tensors.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int,
        recurrent_dropout_probability: float = 0.0,
        use_highway: bool = True,
        use_input_projection_bias: bool = True,
    ) -> None:
        super().__init__()

        # Required to be wrapped with a :class:`PytorchSeq2SeqWrapper`.
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers

        layers = []
        lstm_input_size = input_size
        for layer_index in range(num_layers):
            go_forward = layer_index % 2 == 0
            layer = AugmentedLstm(
                lstm_input_size,
                hidden_size,
                go_forward,
                recurrent_dropout_probability=recurrent_dropout_probability,
                use_highway=use_highway,
                use_input_projection_bias=use_input_projection_bias,
            )
            lstm_input_size = hidden_size
            self.add_module("layer_{}".format(layer_index), layer)
            layers.append(layer)
        self.lstm_layers = layers

    def forward(
        self, inputs: PackedSequence, initial_state: Optional[TensorPair] = None
    ) -> Tuple[Union[torch.Tensor, PackedSequence], TensorPair]:
        """
        Parameters
        ----------
        inputs : ``PackedSequence``, required.
            A batch first ``PackedSequence`` to run the stacked LSTM over.
        initial_state : Tuple[torch.Tensor, torch.Tensor], optional, (default = None)
            A tuple (state, memory) representing the initial hidden state and memory
            of the LSTM. Each tensor has shape (1, batch_size, output_dimension).

        Returns
        -------
        output_sequence : PackedSequence
            The encoded sequence of shape (batch_size, sequence_length, hidden_size)
        final_states: Tuple[torch.Tensor, torch.Tensor]
            The per-layer final (state, memory) states of the LSTM, each with shape
            (num_layers, batch_size, hidden_size).
        """
        if not initial_state:
            hidden_states: List[Optional[TensorPair]] = [None] * len(self.lstm_layers)
        elif initial_state[0].size()[0] != len(self.lstm_layers):
            raise ConfigurationError(
                "Initial states were passed to forward() but the number of "
                "initial states does not match the number of layers."
            )
        else:
            hidden_states = list(zip(initial_state[0].split(1, 0), initial_state[1].split(1, 0)))

        output_sequence = inputs
        final_states = []
        t1 = time.time()
        for i, state in enumerate(hidden_states):
            layer = getattr(self, "layer_{}".format(i))
            # The state is duplicated to mirror the Pytorch API for LSTMs.
            output_sequence, final_state = layer(output_sequence, state)
            final_states.append(final_state)
        final_hidden_state, final_cell_state = tuple(
            torch.cat(state_list, 0) for state_list in zip(*final_states)
        )
        t2 = time.time()
        print("CPU LSTM time: ", t2-t1)

        return output_sequence, (final_hidden_state, final_cell_state)




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import collector
import logging
import json
from tempfile import TemporaryDirectory


def trans_col(i):
    result = {}
    for rec in i.split(';'):
        if len(rec.split(':')) < 2:
            continue
        k, v = rec.split(':', 1)
        result.update({k: v})

    return result


"""
Trace data is stored in dict & sorted by classname
{
    "null" : [...] (classname not specified in raw trace)
    "class_a": [...],
    "class_b": [...]
}
"""


def trans_row(trace_data, xmodel_info):
    current_section = "null"
    result = []

    for d in trace_data:
        if d.startswith('#'):
            if d.startswith("#SECTION"):
                current_section = trans_col(d).get('#SECTION', "null")
            continue

        item = trans_col(d)
        item.update({"section": current_section})
        cname = item.get("classname", "null")

        result.append(item)

    for d in xmodel_info:
        item = json.loads(d)
        item.pop("mc_code_sstr.str()")
        item.update({"section": "XMODEL"})

        result.append(item)

    return result


class tracepointCollector(collector.collectorBase.Collector):
    def __init__(self, stdout="", stderr=""):
        super().__init__(name='tracepoint')
        self.logdir = TemporaryDirectory()
        self.logdirName = self.logdir.name + '/'

    def __del__(self):
        self.logdir.cleanup()

    def prepare(self, conf: dict) -> dict:

        # "VAI_TRACE_ENABLE"=1
        # "VAI_TRACE_DIR"="/tmp/vai/"
        # "VAI_TRACE_TS" = "boot" or "x86_tsc"

        os.environ.setdefault("VAI_TRACE_ENABLE", "1")
        os.environ.setdefault("VAI_TRACE_DIR", self.logdirName)

        # preferClocks = ["boot", "x86-tsc", "global"]
        ts_mode = conf.get("control").get("traceClock")
        logging.debug(ts_mode)
        """
        +--------------+-------------------+------------+
        | clock_source | x86               | arm        |
        +--------------+-------------------+------------+
        | VA           | global + XRT      | boot + XRT |
        | XAT          | x86-tsc + x86-tsc | boot + XRT |
        +--------------+-------------------+------------+
        """
        os.environ.setdefault("VAI_TRACE_TS", ts_mode)

        va_enabled = conf['cmdline_args']['va']
        if va_enabled:
            os.environ["VAI_TRACE_TS"] = "XRT"
            if ts_mode == "x86-tsc":
                """VTF do not accept x86-tsc"""
                conf["control"]["traceClock"] = "global"
                print(conf["control"]["traceClock"])

        return conf

    def start(self):
        super().start()

    def stop(self):
        """Clear env"""
        os.environ.pop("VAI_TRACE_ENABLE", 0)
        super().stop()

    def getData(self):
        traceEvents = []
        xmodelInfo = []

        for root, dirs, files in os.walk(self.logdirName):
            for name in files:
                f = os.path.join(root, name)
                if name.find("subgraph_info") > 0:
                    xmodelInfo += open(f).readlines()
                else:
                    traceEvents += open(f).readlines()

        return trans_row(traceEvents, xmodelInfo)


collector.collectorBase.register(tracepointCollector())




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import ctypes

libvart_util = ctypes.CDLL('libvart-trace.so')
f_tracepoint = libvart_util.tracepoint_py_func
f_tracepoint.argtypes = [ctypes.c_int, ctypes.c_char_p]
f_tracepoint.restype = None


def vai_tracepoint(func):
    @functools.wraps(func)
    def wrapper(*args, **kw):
        fname = func.__name__
        fname_c_char_p = ctypes.c_char_p(bytes(fname, 'ascii'))

        f_tracepoint(1, fname_c_char_p)
        result = func(*args, **kw)
        f_tracepoint(0, fname_c_char_p)

        return result
    return wrapper



"""
Copyright 2019 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from ctypes import *
import numpy as np
import json
import os, sys
import re
import threading

class Tensor(Structure):
  _fields_ = [
    ('name', c_char_p),
    ('dims', POINTER(c_int32)),
    ('ndims', c_int32),
    ('dtype', c_int32)
  ]

class Runner:
  # tensor format enum
  TensorFormat = type('', (), {})()
  TensorFormat.NCHW = 0
  TensorFormat.NHWC = 1

  def __init__(self, path):
    metaFile = os.path.join(path, "meta.json")
    if not os.path.isfile(metaFile):
      raise AssertionError("meta.json file %s not found" % metaFile)
      
    # select .so file based on path/meta.json
    with open(metaFile) as f:
      meta = json.load(f)
      libFile = self._parse_path(meta['lib'])

    if not libFile or not os.path.isfile(libFile):
      raise AssertionError("C++ library .so file %s not found" % libFile)

    self._libFile = os.path.abspath(libFile)
    self._lib = cdll.LoadLibrary(self._libFile)
    self._inputCptrs = {}
    self._outputCptrs = {}

    self._lib.DpuPyRunnerCreate.argtypes = [c_char_p]
    self._lib.DpuPyRunnerCreate.restype = c_void_p
    self._lib.DpuPyRunnerGetInputTensors.argtypes = [c_void_p,
      POINTER(c_void_p), POINTER(c_int)]
    self._lib.DpuPyRunnerGetOutputTensors.argtypes = [c_void_p, 
      POINTER(c_void_p), POINTER(c_int)]
    self._lib.DpuPyRunnerGetTensorFormat.argtypes = [c_void_p]
    self._lib.DpuPyRunnerGetTensorFormat.restype = c_int
    self._lib.DpuPyRunnerExecuteAsync.argtypes = [c_void_p, 
      POINTER(np.ctypeslib.ndpointer(c_float, flags="C_CONTIGUOUS")),
      POINTER(np.ctypeslib.ndpointer(c_float, flags="C_CONTIGUOUS")),
      c_int, POINTER(c_int)]
    self._lib.DpuPyRunnerExecuteAsync.restype = c_int
    self._lib.DpuPyRunnerWait.argtypes = [c_void_p, c_int]
    self._lib.DpuPyRunnerWait.restype = c_int
    self._lib.DpuPyRunnerDestroy.argtypes = [c_void_p]

    self._runner = self._lib.DpuPyRunnerCreate(path.encode('utf-8'))

  def get_input_tensors(self):
    ptr = c_void_p()
    n = c_int(0)
    self._lib.DpuPyRunnerGetInputTensors(self._runner, byref(ptr), byref(n))
    tensors = []
    for i in range(n.value):
      tensors.append(Tensor.from_address(ptr.value + (i*sizeof(Tensor))))
    return tensors

  def get_output_tensors(self):
    ptr = c_void_p()
    n = c_int(0)
    self._lib.DpuPyRunnerGetOutputTensors(self._runner, byref(ptr), byref(n))
    tensors = []
    for i in range(n.value):
      tensors.append(Tensor.from_address(ptr.value + (i*sizeof(Tensor))))
    return tensors

  def get_tensor_format(self):
    return self._lib.DpuPyRunnerGetTensorFormat(self._runner)

  def execute_async(self, inputs, outputs):
    """
      Args:
        inputs: list of numpy arrays
        outputs: list of numpy arrays

        order of numpy arrays in inputs/outputs must match 
          the order in get_input_tensors() and get_output_tensors()
    """
    status = c_int(0)

    # NOTE: every thread needs different input/output pointers
    ret = self._lib.DpuPyRunnerExecuteAsync(self._runner, 
			self._numpy_list_2_cptr_list(inputs, self._inputCptrs), 
      self._numpy_list_2_cptr_list(outputs, self._outputCptrs), 
      inputs[0].shape[0], byref(status))

    if status.value != 0:
      raise RuntimeError("Runner.execute_async could not enqueue new DPU job")

    return ret

  def _numpy_list_2_cptr_list(self, nplist, ptrlist):
    tid = threading.get_ident()
    if tid not in ptrlist or len(nplist) != len(ptrlist[tid]):
      ptrlist[tid] = (np.ctypeslib.ndpointer(c_float, flags="C_CONTIGUOUS") * len(nplist))()
    for i, tensor in enumerate(nplist):
      ptrlist[tid][i] = tensor.ctypes.data

    return ptrlist[tid]

  def _parse_path(self, path):
    """
      Translate any {STRING} in 'path' to os.environ["STRING"]
      E.g., {XILINX_ROOT}/path/to/file to /opt/xilinx/path/to/file
    """
    retpath = path
    regex = r"\{(.*?)\}"
    matches = re.finditer(regex, path, re.MULTILINE | re.DOTALL)
    for matchNum, match in enumerate(matches):
      word = match.group(1)
      retpath = retpath.replace("{"+word+"}", os.environ[word])

    return retpath

  def wait(self, job_id):
    return self._lib.DpuPyRunnerWait(self._runner, job_id)

  def __del__(self):
    if hasattr(self, '_lib') and self._lib \
      and hasattr(self, '_runner') and self._runner:
      self._lib.DpuPyRunnerDestroy(self._runner)



"""
Copyright 2021 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import concurrent.futures
import os
from os import environ, getenv
import json
import numpy as np
import pandas
import tensorflow as tf
import datetime
from keras.datasets import imdb
from keras.models import Sequential
from keras.layers import Dense
from keras.preprocessing import sequence
from keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.models import load_model
from nltk import word_tokenize
from tensorflow.python.keras import Model
import time
import vart
import xir

from argparse import ArgumentParser

parser = ArgumentParser()
parser.add_argument('-n', "--num-runners", default=4, type=int)
args = parser.parse_args()

current_dir = os.path.dirname(__file__)
model_filename = os.path.join(current_dir, "data", "LSTM.h5")
data_filename = os.path.join(current_dir, "data", "imdb.npz")
word_dict_path = os.path.join(current_dir, "data", "imdb_word_index.json")
predict_file = os.path.join(current_dir, "data", "IMDB.csv")
output_predict_file = os.path.join(current_dir, "data", 'predictions.txt')
# set the seed
np.random.seed(7)

# load the dataset top n words only
top_words = 5000
max_review_length = 500
Y = []


def quanti_convert_float_to_int16(data, fix_pos):
    amp = 2**fix_pos
    max = 2**(16-1)

    output = data * amp
    output = np.clip(output, -max, max - 1)
    output = np.floor(output)
    output = output.astype(np.int16)
    return output


def quanti_convert_int16_to_float(data, fix_pos):
    amp = 2**fix_pos
    output = data.astype(np.float32)
    output = data / amp
    return output


def preprocessing(filename, maxlen, top_words, word_dict_path=None):
    noisy_char = '!\"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n'
    table = str.maketrans(noisy_char, ' '*len(noisy_char))
    data = pandas.read_csv(filename, sep=',')
    data.head()
    word_dict = json.load(open(word_dict_path, 'r'))
    X = []

    for sentiment in data['sentiment'].values:
        Y.append(sentiment)
    for review in data['review'].values:
        review = review.translate(table).replace("  ", " ")
        seq = []
        for word in word_tokenize(review):
            if word.lower() not in word_dict:
                continue
            index = word_dict[word.lower()] + 3
            if index >= top_words:
                continue
            seq.append(index)
        X.append(seq)

    X = pad_sequences(X, maxlen=maxlen)
    tf.logging.info("input shape: {}".format(X.shape))

    return X


print("Load model")
model = load_model(model_filename, compile=False)
print("Model ready")
model.summary()
if not os.path.exists(word_dict_path):
    imdb.get_word_index(path=word_dict_path)
print("start preprocessing")
t1 = time.time()
pre_begin = datetime.datetime.now()
# X_test = preprocessing(predict_file, max_review_length,
#                        top_words, word_dict_path)
########################################
is_train = False
# writing the train model and getting input data
if environ.get('RESULT_DIR') is not None:
    output_model_folder = os.path.join(os.environ["RESULT_DIR"], "model")
    output_model_path = os.path.join(output_model_folder, model_filename)
else:
    output_model_folder = "model_bak"
    output_model_path = os.path.join("model", model_filename)

os.makedirs(output_model_folder, exist_ok=True)

if environ.get('DATA_DIR') is not None:
    data_folder = os.environ["DATA_DIR"]
    data_path = os.path.join(data_folder, data_filename)
else:
    data_folder = os.getcwd()
    data_path = os.path.join(data_folder, data_filename)

# set the seed
np.random.seed(7)

# load the dataset top n words only
top_words = 5000
# (X_train, y_train), (X_test, y_test) = load_data(path=data_path,
#                                                  num_words=top_words)

(X_train, y_train), (X_test, y_test) = imdb.load_data(
    path=data_path, num_words=top_words)
# truncate and pad input sequences
max_review_length = 500
X_train = sequence.pad_sequences(X_train, maxlen=max_review_length)
X_test = sequence.pad_sequences(X_test, maxlen=max_review_length)
print("preprocessing over")
pre_end = datetime.datetime.now()
a = np.array(y_test)
a = a.reshape([a.shape[0], 1])


print("Init Dense model")
lstm_model2 = Sequential()
lstm_model2.add(Dense(1, activation='sigmoid'))
lstm_model2.build((1, 100))
layer_dict_fix = dict([(layer.name, layer) for layer in model.layers])
lstm_model2.layers[0].set_weights(layer_dict_fix['dense'].get_weights())
print("Dense ready")
print("Init Hardware")
# model_lstm = dpu4rnn_py.dpu4rnn.create("sentiment")

# Generate rnn graph
embedding_vecor_length = 32
runners = []
batches = []

num_sequences = max_review_length
output_seq_dim = 100

device_name = getenv("TARGET_DEVICE", default="").upper()
assert(device_name in ("U50LV", "U25")), "TARGET_DEVICE should be U50LV/U25"

num_cores = 1
if device_name == "U50LV":
    num_cores = 2
    models = [os.path.join(current_dir, "data", file)
              for file in ["compiled_batch_3.xmodel", "compiled_batch_4.xmodel"]]
elif device_name == "U25":
    num_cores = 1
    models = [os.path.join(current_dir, "data", file)
              for file in ["compiled_batch_1.xmodel"]]

# Setup the Runners
for runner_idx in range(args.num_runners):
    graph = xir.Graph.deserialize(models[runner_idx % num_cores])
    runners.append(vart.Runner.create_runner(graph.get_root_subgraph(), "run"))
    inputTensors = runners[runner_idx].get_input_tensors()
    batch_size = tuple(inputTensors[0].dims)[0]
    batches.append(batch_size)

out_pos = graph.get_root_subgraph().get_attr('output_fix2float')
in_pos = graph.get_root_subgraph().get_attr('input_float2fix')

print("Hardware ready")
hybrid_begin = datetime.datetime.now()

# print(model.get_weights())
print("Embedding start")
permute_layer_model = Model(
    inputs=model.input, outputs=model.get_layer("embedding").output)
# print(permute_layer_model.get_layer("embedding").get_weights())
ebd_begin = datetime.datetime.now()
lstm_input = permute_layer_model.predict(X_test, batch_size=8)
ebd_end = datetime.datetime.now()
num_records = lstm_input.shape[0]
print("Embedding over")

cv1b = datetime.datetime.now()
quantized_lstm_input = quanti_convert_float_to_int16(lstm_input.reshape(
    num_records * 16000), in_pos).reshape((num_records, 16000))
cv1e = datetime.datetime.now()
inputs = []
lstm_output = np.zeros((quantized_lstm_input.shape[0], 100), dtype=np.int16)
print("Start LSTMlayer")
begin = datetime.datetime.now()

# use the multi-batch
inputTensors = runners[0].get_input_tensors()
outputTensors = runners[0].get_output_tensors()
_, _, runner_in_seq_len = tuple(inputTensors[0].dims)
_, _, runner_out_seq_len = tuple(outputTensors[0].dims)

quantized_input = quantized_lstm_input.view()
out_np = lstm_output.view()

def run(input_data, batch_size, count, num_sequences, runner_out_seq_len,
        output_seq_dim, runner, out_np):
    # print("[INFO] Filling {} -> {}".format(count, count+batch_size))
    input_data = input_data.reshape(batch_size, num_sequences, runner_in_seq_len)
    output_data = np.empty((batch_size, num_sequences, runner_out_seq_len), dtype=np.int16)
    job_id = runner.execute_async([input_data], [output_data], True)
    runner.wait(job_id)
    out_np[count:count+batch_size, ...] = output_data[:, -1, :output_seq_dim].reshape(
            batch_size, output_seq_dim)

runner_idx = count = 0
futures = []
print("[INFO] Total number of records: {}".format(num_records))
dpu_start = time.time()
with concurrent.futures.ThreadPoolExecutor() as executor:
    while count < len(quantized_input):
        batch_size = batches[runner_idx]
        input_data = quantized_input[count:count+batch_size]
        batch_size = input_data.shape[0]
        futures.append(executor.submit(
            run, input_data, batch_size, count, num_sequences,
            runner_out_seq_len, output_seq_dim, runners[runner_idx], out_np))
        count += batch_size
        runner_idx = (runner_idx + 1) % args.num_runners

concurrent.futures.wait(futures)
print("[INFO] DPU Time: {}".format(time.time() - dpu_start))

while runners:
    del runners[0]

end = datetime.datetime.now()
print("LSTM over")
cv2b = datetime.datetime.now()
lstmlayerout = quanti_convert_int16_to_float(lstm_output, out_pos)
cv2e = datetime.datetime.now()

dense_begin = datetime.datetime.now()
lstm_output_dense = lstm_model2.predict(lstmlayerout, batch_size=8)
dense_end = datetime.datetime.now()
hybrid_end = datetime.datetime.now()
t2 = time.time()
lacc = []

for ind, j in enumerate(lstm_output_dense):
    if (lstm_output_dense[ind] > 0.5):
        lacc.append(1)
    else:
        lacc.append(0)
lacc = np.array(lacc).reshape(len(lacc), 1)
countall = lacc.shape[0]
counterr = 0
for ind, i in enumerate(lacc):
    if (lacc[ind] != a[ind]):
        counterr = counterr + 1
hybrid_end = datetime.datetime.now()

print("====== IMDB Sentiment Detection Results ======")
print("Accuracy: ", (countall - counterr)/countall)
print("preprocessing time", (pre_end - pre_begin).total_seconds())
print("embedding time", (ebd_end - ebd_begin).total_seconds())
print("convert1", (cv1e - cv1b).total_seconds())
print("total hw time", (end - begin).total_seconds())
print("convert2", (cv2e - cv2b).total_seconds())
print("dense lstm time", (dense_end - dense_begin).total_seconds())
print("total hybrid time", (hybrid_end - hybrid_begin).total_seconds())
print("num", out_np.shape[0])
print("Avg hw time", (end - begin).total_seconds() / out_np.shape[0])
print()
print("total hybrid time:", t2-t1)



#Copyright 2021 Xilinx Inc.
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

sents = set()
with open("./utils/test.oie", 'r') as f:
    for line in f:
        sent = line.strip().split("\t")[0]
        sents.add(sent)
with open("./test/test.oie.sent", 'w') as f:
    for sent in sents:
        f.write(sent + "\n")
with open("./test/test_in.txt", 'w') as f:
    f.write(sent)



# Copyright 2021 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import logging
import copy
from ctypes import *
import logging
import binascii


class dpuMcParserBase:
    def __init__(self, _name="unknow"):
        self.name = _name
        self.data = {"load_img_size": 0, "load_para_size": 0,
                     "save_size": 0, "workload": 0}
        self.debug = False

    def compatible(self, dpu_name):
        if dpu_name.startswith(self.name):
            logging.debug("Compatible [%s]: True" % self.name)
            return True

        return False

    def get_inst(self, _mc, pos, dpu_cls):
        head = self.inst_head()
        memmove(addressof(head), _mc[pos:], sizeof(head))

        opcode = head.opcode
        for op in dpu_cls.opcode_table:
            if op.opcode == opcode:
                return (op.name, op.length_byte)

        if self.debug:
            raise ValueError("Unrecognized instruction")

        logging.error("Unrecognized instruction")
        return ("UNKNOW", 0)

    def process(self, mc, _debug=False):

        self.debug = _debug

        pos = 0
        load_img_size = 0
        load_para_size = 0
        save_size = 0

        while (pos < len(mc)):
            inst_name, inst_len = self.get_inst(mc, pos, self)

            if self.debug:
                print("pos: %d/%d, inst: %s          " %
                      (pos, len(mc), inst_name), end="\r")
            if inst_name == "LOAD":
                l = self.Load()
                inst = mc[pos:]
                memmove(addressof(l), inst, inst_len)

                length_ = l.length + 1
                chan_ = l.channel + 1
                try:
                    block_num_ = l.block_num + 1
                except:
                    block_num_ = 1

                #print("LOAD,%d,%d,%d" % (length_, chan_,block_num_))
                load_img_size += (length_ * chan_ * block_num_)

            elif inst_name == "SAVE":
                s = self.Save()
                inst = mc[pos:]
                memmove(addressof(s), inst, inst_len)

                length_ = s.length + 1
                chan_ = s.channel + 1
                save_size += (length_ * chan_)
                #print("SAVE %d %d" % (length_, chan_))
            elif inst_name == "UNKNOW":
                load_img_size = 0
                load_para_size = 0
                save_size = 0
                break
            else:
                pass

            pos += inst_len

        self.data["load_img_size"] = load_img_size
        self.data["load_para_size"] = load_para_size
        self.data["save_size"] = save_size

    """
    by default, return "load_img_size, load_para_size, save_size, workload",
    this method can be overrde
    """

    def get_data(self, items: list) -> list:
        ret = []
        ret.append(self.data.get("load_img_size"))
        ret.append(self.data.get("load_para_size"))
        ret.append(self.data.get("save_size"))
        ret.append(self.data.get("workload"))

        return ret

    class inst_head(LittleEndianStructure):
        _pack_ = 1
        _fields_ = [
            ('resverd', c_uint, 20),
            ('dpby', c_uint, 4),
            ('dpdon', c_uint, 4),
            ('opcode', c_uint, 4)
        ]

    class inst_desc():
        def __init__(self, _name, _opcode, _length_w):
            self.name = _name
            self.opcode = _opcode
            self.length_w = _length_w
            self.length_byte = _length_w * 4


__dpu_mc_parser = []


"""
by default, reture [load_img_size, load_para_size, save_size, workload]
"""


def process_mc(dpu_name: str, mc_data_str: str, _debug=False, ret_items=["load_img_size", "load_para_size", "save_size", "workload"]) -> list:
    # init return data with 0
    ret = []
    for i in range(len(ret_items)):
        ret.append(0)

    # 1. find dpu mc parser
    matched = False
    for parser in __dpu_mc_parser:
        if parser.compatible(dpu_name) == False:
            continue
        matched = True
        logging.debug("Processing DPU mc [%s]..." % parser.name)

        # 2. process_mc
        mc = binascii.a2b_hex(mc_data_str)
        parser.process(mc, _debug)

        # 3. get_data
        ret = parser.get_data(ret_items)

    if matched == False:
        logging.warning(
            "Can't find dpu mc parser for %s, return all 0" % dpu_name)

    return ret


def register(mcParserInstance):
    __dpu_mc_parser.append(mcParserInstance)




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import collector


class fileCollector(collector.collectorBase.Collector):
    def __init__(self):
        super().__init__(name='file')
        self.fileList = []

    def prepare(self, conf: dict) -> dict:
        fileOption = conf.get('collector', {}).get('file')
        for f in fileOption.keys():
            self.fileList.append(f)
            fileInstOption = fileOption[f]

        return conf

    def start(self):
        super().start()

    def stop(self):
        super().stop()

    """
    Return Format:
    {
      'filename': content,
      'filename2': content,
      ...
    }
    """

    def getData(self):
        data = {}

        for f in self.fileList:
            data.update({f: open(f, 'rb').read()})
            print("@#@@@@ open %s" % f)

        return data


collector.collectorBase.register(fileCollector())



#!/usr/bin/python3

# Copyright 2021 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import xir
import dpu_mc_parser
import logging
import binascii
import argparse
import time

PARSE_TIME_ERR = 30
LOG_PATH = "./test_dpu_mc.log"


def find_xmodel(dir):
    suffix = ".xmodel"
    ret = []
    for root, dirs, files in os.walk(dir):
        s = [x for x in files if x.endswith(suffix)]
        for x in s:
            ret.append(os.path.join(root, x))

    return ret


def xmodel_to_subgraphs(xmodel):
    graph = xir.Graph.deserialize(xmodel)
    root = graph.get_root_subgraph()

    # add subgraphs of depth=1
    ret = list(root.toposort_child_subgraph())

    # add subgraphs of depth=2
    for dp1_subgraph in root.toposort_child_subgraph():
        for dp2_subgraph in dp1_subgraph.toposort_child_subgraph():
            ret.append(dp2_subgraph)

    return root.toposort_child_subgraph()


def subgraph_to_mc_str(subg):
    name = subg.get_name()

    if subg.has_attr("mc_code"):
        mc = subg.get_attr("mc_code")
        mc_str = binascii.b2a_hex(mc)
        return mc_str

    return None


def process_mc_str(dpu_name, mc_str):
    mc_data = dpu_mc_parser.process_mc(dpu_name, mc_str, True)
    load_img_size = mc_data[0]
    load_para_size = mc_data[1]
    save_size = mc_data[2]
    workload = mc_data[3]

    return load_img_size, load_para_size, save_size, workload


if __name__ == "__main__":
    # process logging
    logging.basicConfig(level=logging.INFO)
    os.system("echo > %s" % LOG_PATH)
    fhlr = logging.FileHandler(LOG_PATH)
    logging.getLogger().addHandler(fhlr)

    # process args
    cmd_parser = argparse.ArgumentParser(
        usage="Usage: ./test_dpu_mc.py <xmodel_dir_path> -d <depth>\ne.g: ./writer/parser/test_dpu_mc.py /usr/share/vitis_ai_library/models/ -d 1")
    cmd_parser.add_argument('-d', '--depth', default=1,
                            choices=[1, 2], type=int, help="xmodel parse depth, default: 1")
    cmd_parser.add_argument("xmodel_dir", nargs=1, help="path to xmodel dir")
    args = cmd_parser.parse_args()

    xm_dir = args.xmodel_dir[0]
    xmodel_parse_depth = args.depth

    if not os.path.isdir(xm_dir):
        raise RuntimeError("Invalid directory")

    xmodels = find_xmodel(xm_dir)
    n_models = len(xmodels)

    logging.info("%s%s%s" % (50 * "=", "[START]", 50 * "="))
    for m in xmodels:
        subgs = xmodel_to_subgraphs(m)
        logging.info("Processing xmodel [%d/%d]: %s, # of subgraphs: %d" %
                     (xmodels.index(m) + 1, n_models, m, len(subgs)))
        logging_prefix = 4 * " "

        for subg in subgs:
            t_subgraph_start = time.time()
            if (subg.has_attr("device")):
                device = subg.get_attr("device")
                if device != "DPU":
                    continue

            if (subg.has_attr("dpu_name")):
                dpu_name = subg.get_attr("dpu_name")
            else:
                logging.error("Cannot find <dpu_name> in xmodel")

            # process depth=1 subgraphs
            mc_str = subgraph_to_mc_str(subg)
            if mc_str != None:
                logging.info(logging_prefix + "Processing subgraph: %s, dpu_name: %s" %
                             (subg.get_name(), dpu_name))
                load_img_size, load_para_size, save_size, workload = process_mc_str(
                    dpu_name, mc_str)
                if workload == 0:
                    """Try to get workload from xmodel"""
                    if subg.has_attr("workload"):
                        workload = subg.get_attr("workload")
                logging.info(logging_prefix + "Result: LoadImgSize: {:,}, LoadParaSize: {:,}, SaveSize: {:,}, Workload: {:,}".format
                             (load_img_size, load_para_size, save_size, workload))
                t_subgraph_dur = (time.time() - t_subgraph_start)
                if t_subgraph_dur < PARSE_TIME_ERR:
                    logging.info(
                        logging_prefix + "Parsing time consuming: {:.3f} s, mc_size: {:,}".format(t_subgraph_dur, len(mc_str)))
                else:
                    logging.error(
                        logging_prefix + "ERROR: Parsing time consuming: {:.3f} s, mc_size: {:,}".format(t_subgraph_dur, len(mc_str)))
                logging.info(logging_prefix + 80 * "-")

            # process depth=1 subgraphs
            if xmodel_parse_depth == 1:
                continue
            for dp2_subg in subg.toposort_child_subgraph():
                mc_str = subgraph_to_mc_str(dp2_subg)
                if mc_str != None:
                    logging.info(logging_prefix + "Processing subgraph of depth=2: %s, dpu_name: %s" %
                                 (dp2_subg.get_name(), dpu_name))
                    load_img_size, load_para_size, save_size, workload = process_mc_str(
                        dpu_name, mc_str)

                    if workload == 0:
                        """Try to get workload from xmodel"""
                        if dp2_subg.has_attr("workload"):
                            workload = dp2_subg.get_attr("workload")
                    logging.info(logging_prefix + "Result: LoadImgSize: {:,}, LoadParaSize: {:,}, SaveSize: {:,}, Workload: {:,}".format
                                 (load_img_size, load_para_size, save_size, workload))
                    logging.info(logging_prefix + 80 * "-")
        logging.info(80 * "-")

    logging.info("%s%s%s" % (50 * "=", "[END]", 50 * "="))
    logging.warning("Log File: [%s]" % os.path.abspath(LOG_PATH))




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tracer.tracerBase
import sys
import os
import ctypes
import logging

HW_zynq = [
    {'type': 'CPU', 'file': ['/proc/cpuinfo',
                             '/sys/kernel/debug/clk/acpu/clk_rate']},
    {'type': 'Memory', 'file': ['/proc/meminfo']}
]

HW_alveo = [
    {'type': 'CPU', 'file': ['/proc/cpuinfo']},
]
"""   IP_LAYOUT SECTION   """


class IP_CONTROL:
    AP_CTRL_HS = 0
    AP_CTRL_CHAIN = 1
    AP_CTRL_NONE = 2
    AP_CTRL_ME = 3


class indices (ctypes.Structure):
    _fields_ = [
        ("m_index", ctypes.c_uint16),
        ("m_pc_index", ctypes.c_uint8),
        ("unused", ctypes.c_uint8)
    ]


class ip_u1 (ctypes.Union):
    _fields_ = [
        ("m_base_address", ctypes.c_int64),
        ("indices", indices)
    ]


class ip_data (ctypes.Structure):
    _fields_ = [
        ("m_type", ctypes.c_uint32),
        ("properties", ctypes.c_uint32),
        ("ip_u1", ip_u1),
        ("m_name", ctypes.c_uint8 * 64)
    ]


class ip_layout (ctypes.Structure):
    _fields_ = [
        ("m_count", ctypes.c_int32),
        ("m_ip_data", ip_data*16)
    ]


def parseText(_type, text, platform):
    cur = dict()
    if _type == 'CPU':
        if platform['machine'] == 'x86_64':
            cur['Cores'] = len([t for t in text if t.startswith('processor')])
            cur['Model'] = "%s" % [t for t in text if t.startswith(
                'model name')][0].split(':')[1].strip()
        else:
            cur['Cores'] = len([t for t in text if t.startswith('processor')])
            cur['Model'] = "ARM-v%s" % [t for t in text if t.startswith(
                'CPU architecture')][0].split(':')[1].strip()

            for l in text:
                if l.find('acpu') >= 0:
                    freq = text[(text.index(l)+1)].strip()
                    cur['Freq'] = "%d MHz" % (int(freq) / 1000 / 1000)
                    break

    if _type == 'Memory':
        totalMem = int([t for t in text if t.startswith('MemTotal')]
                       [0].split(':')[1].strip().split(' ')[0])
        cur['Total Memory'] = "%d MB" % (totalMem / 1024)

    if _type == 'DPU':
        core_id = None

        for t in text:
            ts = t.split(':')
            if len(ts) != 2:
                continue

            key = ts[0].strip()
            value = ts[1].strip()
            if key == 'DPU Core':
                core_id = "%s_%s" % (_type, value)
                cur.setdefault(core_id, {})
            else:
                if core_id == None:
                    continue
                cur[core_id].setdefault(key, value)

    return {'type': _type, 'info': cur}


def getIPsZynq(path=['/sys/devices/platform/amba/amba:zyxclmm_drm/ip_layout',
                     '/sys/devices/platform/axi/axi:zyxclmm_drm/ip_layout',
                     '/sys/devices/platform/amba_pl@0/amba_pl@0:zyxclmm_drm/ip_layout']):
    cus = {}

    for p in path:
        try:
            ip_layout_data = open(p, 'rb').read()
            ips = ip_layout()

            ctypes.memmove(ctypes.pointer(
                ips), ip_layout_data, len(ip_layout_data))

            for i in range(0, ips.m_count):
                ip = ips.m_ip_data[i]
                cu_name = str()

                for c in ip.m_name:
                    cu_name += chr(c)
                cu_type, cu_name = cu_name.strip('\x00').split(':')

                cu_name = cu_name.replace('_xrt_top', '')
                cu_paddr = hex(ip.ip_u1.m_base_address)

                cus[cu_name] = cu_paddr
            if len(cus) > 0:
                break
        except:
            pass

    if len(cus) == 0:
        logging.warning("Cannot open 'zyxclmm_drm/ip_layout'")

    return {'type': 'CUMap', 'info': cus}


def getHwInfo(platform):
    hwInfo = []
    machine = platform.get('machine', "")
    if machine == 'x86_64':
        hwlist = HW_alveo
    elif machine == 'aarch64':
        hwlist = HW_zynq
    else:
        logging.warning("Un-supported platform")
        return hwInfo

    for h in hwlist:
        text = []
        for f in h['file']:
            """Read method for a file"""
            try:
                if os.path.exists(f):
                    text.append("[%s]\n" % f)
                    text += open(f, 'r').readlines()
                else:
                    text += os.popen(f)
            except:
                logging.error("Getting Hardware Info from [%s] failed" % h)

        hwInfo.append(parseText(h['type'], text, platform))

    if machine == 'aarch64':
        hwInfo.append(getIPsZynq())

    return hwInfo


class hwInfoTracer(tracer.tracerBase.Tracer):
    def __init__(self):
        super().__init__('hwInfo', source=[], compatible={
            'machine': ["aarch64", "x86_64"]})

    def prepare(self, options: dict, debug: bool):
        self.platform = options.get('control').get('platform')

    def process(self, data, t_range=[]):
        pass

    def getData(self):
        return getHwInfo(self.platform)


tracer.tracerBase.register(hwInfoTracer())




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from shutil import copy

setup(
    name='vaitrace_py',
    version='1.3',
    description="",
    py_modules=["__init__", "vaitrace_py", "tracepoint"],
    python_requires='>=3.5',
    zip_safe=False,
)



"""
Copyright 2021 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pdb
import time
import argparse
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import vart
import xir
from tensorflow import keras
from tensorflow.keras.datasets import imdb
from tensorflow.keras.preprocessing import sequence
from torch.utils.data import DataLoader, TensorDataset
from threading import Thread

DEVICE = torch.device("cpu")

top_words = 5000  # vocab size
max_review_length = 500
embedding_vector_length = 32
total_vec_length = max_review_length * embedding_vector_length
hidden_vector_length = 100


class PreProcModel(nn.Module):
    def __init__(self, max_words, emb_size, hid_size):
        super(PreProcModel, self).__init__()
        self.max_words = max_words
        self.emb_size = emb_size
        self.hid_size = hid_size
        self.Embedding = nn.Embedding(self.max_words, self.emb_size)

    def forward(self, x):
        print("Running Preproc ...")
        return self.Embedding(x)


class GRUModel(nn.Module):
    def __init__(self, max_words, emb_size, hid_size):
        super(GRUModel, self).__init__()
        self.max_words = max_words
        self.emb_size = emb_size
        self.hid_size = hid_size
        self.GRU = nn.GRU(self.emb_size, self.hid_size, batch_first=True)

    def forward(self, x):
        print("Running GRU on CPU ...")
        y, _ = self.GRU(x)
        z = y[:, -1, :]
        return z


class PostProcModel(nn.Module):
    def __init__(self, max_words, emb_size, hid_size):
        super(PostProcModel, self).__init__()
        self.max_words = max_words
        self.emb_size = emb_size
        self.hid_size = hid_size
        self.fc = nn.Linear(self.hid_size, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        print("Running Postproc ...")
        #  x = x[:, -1, :]
        x = self.fc(x)
        out = self.sigmoid(x)
        return out.view(-1)

def run_impl(in_np, begin, end, num_sequences, output_seq_dim, runner, out_np):
    start_id = begin
    inputTensors = runner.get_input_tensors()
    outputTensors = runner.get_output_tensors()
    batch_size, num_frames, runner_in_seq_len = tuple(inputTensors[0].dims)
    _, _, runner_out_seq_len = tuple(outputTensors[0].dims)
    output_buf = np.empty((end - start_id, 1, output_seq_dim), dtype=np.int16)

    while(begin < end):
        eff_batch = min(end-begin, batch_size)
        input_data = in_np[begin: begin + eff_batch].reshape(-1, num_sequences, runner_in_seq_len)
        output_data = np.empty((eff_batch, num_sequences, runner_out_seq_len), dtype=np.int16)
        job_id = runner.execute_async([input_data], [output_data], True)
        runner.wait(job_id)
        offset = begin - start_id
        output_buf[offset:offset+eff_batch, ...] = output_data[:, -1, :output_seq_dim].reshape(
                  eff_batch, 1, output_seq_dim)
        begin += eff_batch
    out_np[start_id:end, ...] = output_buf


class HWRun:
    def __init__(self, num_cores=2, num_sequences=max_review_length, output_seq_dim=hidden_vector_length):
        # Setup the Runners
        self.num_sequences = num_sequences
        self.output_seq_dim = output_seq_dim
        self.num_cores = num_cores
        models = [
            "data/compiled_batch_3.xmodel",
            "data/compiled_batch_4.xmodel",
        ]
        self.runners = []
        self.batches = []
        for i in range(self.num_cores):
            model = models[i % len(models)]
            print("Initiating Runner-{} from {}".format(i, model))
            graph = xir.Graph.deserialize(model)
            self.runners.append(vart.Runner.create_runner(graph.get_root_subgraph(), "run"))
            inputTensors = self.runners[i].get_input_tensors()
            outputTensors = self.runners[i].get_output_tensors()
            batch_size, num_frames, runner_in_seq_len = tuple(inputTensors[0].dims)
            _, _, runner_out_seq_len = tuple(outputTensors[0].dims)
            self.batches.append(batch_size)
        self.out_pos = graph.get_root_subgraph().get_attr("output_fix2float")
        self.in_pos = graph.get_root_subgraph().get_attr("input_float2fix")

    def run(self, xin_pt):
        print("Running GRU on DPU ...")
        inputTensors = self.runners[0].get_input_tensors()
        outputTensors = self.runners[0].get_output_tensors()
        batch_size, num_frames, runner_in_seq_len = tuple(inputTensors[0].dims)
        _, _, runner_out_seq_len = tuple(outputTensors[0].dims)
        xin_np = xin_pt.cpu().numpy()
        num_records = xin_np.shape[0]
        t1 = time.time()
        quantized_lstm_input = quanti_convert_float_to_int16(xin_np, self.in_pos).reshape(
            (num_records, total_vec_length)
        )
        t2 = time.time()

        lstm_output = np.zeros((quantized_lstm_input.shape[0], 1, self.output_seq_dim), dtype=np.int16)

        quantized_input = quantized_lstm_input.view()
        out_np = lstm_output.view()
        threads = []
        begin = 0
        nrec = math.ceil(num_records / sum(self.batches))
        for i, runner in enumerate(self.runners):
            end = min(num_records, begin + nrec * self.batches[i])
            print("Runner : {}, Batch Size : {}, #Batches : {}".format(
                i, self.batches[i], math.ceil((end-begin)/self.batches[i]))
            )
            threads.append(Thread(target=run_impl,
                                  args=(quantized_input, begin, end,
                                        self.num_sequences, self.output_seq_dim,
                                        runner, out_np)))
            begin = end

        t3 = time.time()
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        t4 = time.time()

        lstmlayerout = quanti_convert_int16_to_float(lstm_output, self.out_pos)
        t5 = time.time()
        yout_pt = torch.from_numpy(lstmlayerout).cpu()
        return yout_pt, ((t2-t1)*1000, (t4-t3)*1000, (t5-t4)*1000)

    def __del__(self):
        while self.runners:
            del self.runners[0]


def test(preprocModel, postprocModel, gruModel, hwModel, test_loader, device="CPU"):
    preprocModel.eval()
    gruModel.eval()
    postprocModel.eval()
    #  criterion = nn.BCELoss()
    acc = 0
    for batch_idx, (x, y) in enumerate(test_loader):
        x, y = x.to(DEVICE), y.to(DEVICE)
        with torch.no_grad():
            t1 = time.time()
            pre_y_ = preprocModel(x)
            t2 = time.time()
            if(device == "CPU"):
                gru_y_ = gruModel(pre_y_)
            elif(device == "U50LV"):
                gru_y_, (tq, td, tdq) = hwModel.run(pre_y_)
            t3 = time.time()
            y_ = postprocModel(gru_y_)
            t4 = time.time()
        #  test_loss += criterion(y_, y)
        acc += (
            y.cpu().numpy().astype(int) == (y_ > 0.5).cpu().numpy().astype(int)
        ).sum()
    #  test_loss /= len(test_loader.dataset)
    print("Preprocess time : {:.2f} ms".format((t2-t1) * 1000))
    if(device == "CPU"):
        print("GRU (CPU) time : {:.2f} ms".format((t3-t2) * 1000))
    else:
        print("Quantize time : {:.2f} ms".format(tq))
        print("GRU ({}) time : {:.2f} ms".format(device, td))
        print("Dequantize time : {:.2f} ms".format(tdq))
    print("Postprocess time : {:.2f} ms".format((t4-t3) * 1000))
    print("-----------------------------------------")
    print("Total time : {:.2f} ms".format((t4-t1) * 1000))
    print(
        "\nTest set: Accuracy: {}/{} ({:.0f}%)".format(
            acc,
            len(test_loader.dataset),
            100.0 * acc / len(test_loader.dataset),
        )
    )
    return acc / len(test_loader.dataset)


def quanti_convert_float_to_int16(data, fix_pos):
    amp = 2 ** fix_pos
    max = 2 ** (16 - 1)

    output = data * amp
    output = np.clip(output, -max, max - 1)
    #  mask = 0.5 * np.ones_like(output)
    #  mask[output < 0] = -0.5
    mask = 0
    output = np.floor(output + mask)
    output = output.astype(np.int16)
    return output


def quanti_convert_int16_to_float(data, fix_pos):
    amp = 2 ** fix_pos
    output = data.astype(np.float32)
    output = data / amp
    return output.astype(np.float32)


def main(args):
    print("Loading data ...")
    (_, _), (X_test, y_test) = imdb.load_data(path="imdb.npz", num_words=top_words)
    X_test = sequence.pad_sequences(X_test, maxlen=max_review_length)

    test_data = TensorDataset(torch.LongTensor(X_test), torch.FloatTensor(y_test))
    test_loader = DataLoader(test_data, batch_size=25000, shuffle=False)
    PATH = "data/model.best.pth"

    preprocmodel = PreProcModel(
        top_words, embedding_vector_length, hidden_vector_length
    ).to(DEVICE)
    print(preprocmodel)
    preprocmodel.load_state_dict(
        torch.load(PATH, map_location=torch.device("cpu")), strict=False
    )

    grumodel = GRUModel(
        top_words, embedding_vector_length, hidden_vector_length
    ).to(DEVICE)
    print(grumodel)
    grumodel.load_state_dict(
        torch.load(PATH, map_location=torch.device("cpu")), strict=False
    )

    # PostProc
    postprocmodel = PostProcModel(
        top_words, embedding_vector_length, hidden_vector_length
    ).to(DEVICE)
    print(postprocmodel)
    postprocmodel.load_state_dict(
        torch.load(PATH, map_location=torch.device("cpu")), strict=False
    )

    # HWModel
    hwModel = HWRun(args.num_runners) if args.device != "CPU" else None
    #  optimizer = optim.Adam(postprocmodel.parameters())

    _ = test(preprocmodel, postprocmodel, grumodel, hwModel, test_loader, args.device)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--device", required=True,
                        help="specify the device for inference",
                        type=str, choices=("CPU", "U50LV"))
    parser.add_argument("-n", "--num_runners", default=2, type=int,
                        help="number of parallel runners")
    args = parser.parse_args()
    main(args)




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tracer.tracerBase
import sys
import os
import json
import ctypes


class cuEdgeTracer(tracer.tracerBase.Tracer):
    def __init__(self):
        super().__init__('cuEdge', source=[
            "ftrace"], compatible={'machine': ["aarch64"]})

    def prepare(self, conf: dict, debug: bool):
        "Handle Input Options"

        "Handle Output Options"
        optForFtrace = {
            "collector": {
                "ftrace": {
                    "cuEdge": {
                        "name": "cuEdge",
                        "type": "kprobe",
                        "saveTo": './cuEdge.trace',
                        "traceList": [
                            ["cu_start", "zocl", "zocl_hls_start",
                                ["cu_idx=+0(%x0):u32"]],
                            ["cu_done",  "zocl", "zocl_hls_check+0x70",
                                ["cu_idx=+0(%x20):u32"]]
                        ]
                    }
                }
            }
        }

        return optForFtrace

    def process(self, data, t_range):
        tmp = [l for l in data.get('ftrace', {}).get(
            self.name) if not l.startswith('#')]
        self.data = [l for l in tmp if l.find('cu_idx=0') < 0]

    def getData(self):
        return self.data


tracer.tracerBase.register(cuEdgeTracer())



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import csv
import sys
import string
import time
import argparse
import json
import signal
import platform
import gzip
import pickle
import time
import logging
from subprocess import Popen, PIPE

import tracer
import writer
import vaitraceCfgManager
import vaitraceSetting

import writer.vtf.convert
import writer.txt.convert


def dumpRaw(data, saveTo, compress=False):
    if compress:
        xat_file = gzip.open(saveTo, 'wb+')
    else:
        xat_file = open(saveTo, 'wb+')

    pickle.dump(data, xat_file)

    logging.info(".xat file Saved to [%s]" % os.path.abspath(saveTo))
    xat_file.close()


def write(globalOptions: dict):
    options = globalOptions
    debug = options.get('cmdline_args', {}).get('debug', False)

    rawData = tracer.getData()
    outFmt = 'xat'
    va_enabled = options.get('cmdline_args', {}).get('va', False)
    if va_enabled:
        outFmt = 'vtf'

    txt_enabled = options.get('cmdline_args', {}).get('txt', False)
    if txt_enabled:
        """Override outfmt to txt"""
        outFmt = 'txt'

    if (debug):
        dumpRaw(rawData, "./vtf_debug.xat", True)

    if outFmt == 'xat':
        """Dump Data"""
        saveTo = options.get('control').get('xat').get('filename')
        compress = options.get('control').get('xat').get('compress', True)

        dumpRaw(rawData, saveTo, compress)

    elif outFmt == 'vtf':
        logging.info("Generating VTF")
        writer.vtf.convert.xat_to_vtf(rawData)

    elif outFmt == 'txt':
        logging.info("Generating ascii-table summary")
        saveTo = options['cmdline_args']['output']
        writer.txt.convert.xat_to_txt(rawData, saveTo)

    else:
        logging.error("Undefined format")
        exit(-1)


if __name__ == "__main__":
    input_xat = sys.argv[1]
    if len(sys.argv) == 3:
        output_path = sys.argv[2]
    elif len(sys.argv) == 2:
        output_path = os.path.dirname(input_xat)

    f = gzip.open(input_xat, 'rb')
    xat = pickle.load(f)

    #writer.vtf.convert.xat_to_vtf(xat, output_path)
    writer.txt.convert.xat_to_txt(xat)



# Copyright 2021 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
import sys
import os
import logging
import copy
from ctypes import *
import xir

SUBGRAPH_DB = {}
DPU_NAME = ""


class Load(BigEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('opcode', c_uint, 4),
        ('dpdon', c_uint, 4),
        ('dpby', c_uint, 4),
        ('bank_id', c_uint, 6),
        ('bank_addr', c_uint, 14),
        ('r0', c_uint, 1),
        ('pad_start', c_uint, 5),
        ('pad_end', c_uint, 5),
        ('pad_idx', c_uint, 5),
        ('jump_read', c_uint, 16),
        ('jump_write', c_uint, 8),
        ('length', c_uint, 10),
        ('mode_avg', c_uint, 2),
        ('channel', c_uint, 12),
        ('reg_id', c_uint, 3),
        ('ddr_addr', c_uint, 29),
        ('r1', c_uint, 8),
        ('block_num', c_uint, 10),
        ('jump_write_endl', c_uint, 14)
    ]


class Save(BigEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('opcode', c_uint, 4),
        ('dpdon', c_uint, 4),
        ('dpby', c_uint, 4),
        ('bank_id', c_uint, 6),
        ('bank_addr', c_uint, 14),
        ('r0', c_uint, 16),
        ('jump_write', c_uint, 16),
        ('jump_read', c_uint, 8),
        ('length', c_uint, 10),
        ('r1', c_uint, 2),
        ('channel', c_uint, 12),
        ('reg_id', c_uint, 3),
        ('ddr_addr', c_uint, 29)
    ]


opcode_table = [
    {
        'name': "LOAD",
        'opcode': 0b0000,
        'length_w': 5,
        'layout': ""
    }, {
        'name': "SAVE",
        'opcode': 0b0100,
        'length_w': 4,
        'layout': ""
    }, {
        'name': "CONV",
        'opcode': 0b1000,
        'length_w': 5
    }, {

        'name': "CONVINIT",
        'opcode': 0b1001,
        'length_w': 6
    }, {

        'name': "DPTWISE",
        'opcode': 0b1010,
        'length_w': 5
    }, {
        'name': "DWINIT",
        'opcode': 0b1011,
        'length_w': 4
    }, {

        'name': "POOLINIT",
        'opcode': 0b0110,
        'length_w': 2
    }, {

        'name': "POOL",
        'opcode': 0b1100,
        'length_w': 5
    }, {

        'name': "ELEWINIT",
        'opcode': 0b1101,
        'length_w': 2
    }, {

        'name': "ELEW",
        'opcode': 0b1110,
        'length_w': 3
    }, {

        'name': "END",
        'opcode': 0b0111,
        'length_w': 1
    }
]


def process_mc(mc):
    global DPU_NAME
    if not DPU_NAME.startswith("DPUCVDX8G_ISA0"):
        return 0, 0

    pos = 0

    def get_next_op(_mc):
        opcode = _mc[pos+3] >> 4
        load_size = 0
        save_size = 0

        for op in opcode_table:
            if op['opcode'] == opcode:
                mc_name = op['name']
                if mc_name == "LOAD":
                    l = Load()
                    byte_length = op['length_w']
                    for i in range(0, byte_length):
                        for j in range(0, 4):
                            dest_start = addressof(l)
                            dest_off = i * 4 + j
                            src_off = i * 4 + 3 - j
                            src_buf = _mc[pos + src_off]
                            memset(dest_start+dest_off, src_buf, 1)

                    length_ = l.length + 1
                    chan_ = l.channel + 1
                    block_num_ = l.block_num + 1
                    #print("LOAD,%d,%d,%d" % (length_, chan_, block_num_))
                    load_size += (length_ * chan_ * block_num_)

                if mc_name == "SAVE":
                    s = Save()
                    byte_length = op['length_w']
                    for i in range(0, byte_length):
                        for j in range(0, 4):
                            dest_start = addressof(s)
                            dest_off = i * 4 + j
                            src_off = i * 4 + 3 - j
                            src_buf = _mc[pos + src_off]
                            memset(addressof(s)+dest_off, src_buf, 1)

                    length_ = s.length + 1
                    chan_ = s.channel + 1
                    save_size += (length_ * chan_)
                    #print("SAVE %d %d" % (length_, chan_))

                return op['length_w'] * 4, load_size, save_size
        assert(False)

    load_size = 0
    save_size = 0
    while (pos < len(mc)):
        offsit, load_s, save_s = get_next_op(mc)
        pos += offsit
        load_size += load_s
        save_size += save_s
    return load_size, save_size


def to_unit(value, unit="m", precision=3):
    if unit.lower().startswith("k"):
        value = value / 1024
    elif unit.lower().startswith("m"):
        value = value / 1024 / 1024
    elif unit.lower().startswith("g"):
        value = value / 1024 / 1024 / 1024
    else:
        raise("Unit format error")

    return "%%.%df" % precision % value


def find_in_inc(inc, key):
    value = 0
    incs = inc.split()
    for i in incs:
        if i == key:
            idx = incs.index(i)
    value = incs[idx+1]
    return value


def inc_get_io_tensor_size(inc):
    ch = 1
    length = 1
    block_num = 1
    if inc.startswith("LOAD"):
        ch = int(find_in_inc(inc, 'channel'))
        length = int(find_in_inc(inc, "length"))
        block_num = int(find_in_inc(inc, 'block_num'))
    elif inc.startswith("SAVE"):
        ch = int(find_in_inc(inc, 'channel'))
        length = int(find_in_inc(inc, "length"))
    else:
        pass

    return ch * length * block_num


def subg_to_id(sub_g, key="workload"):
    if key == "workload":
        workload = 0
        name = sub_g.get_name()
        if (sub_g.has_attr("workload")):
            workload = sub_g.get_attr("workload")
        return "%s|%d" % (name, workload)
    if key == "depth":
        name = sub_g.get_name()
        depth = sub_g.depth
        return "%s|%d" % (name, depth)


def idx_to_name(idx):
    return idx.rsplit('|', 2)[0]


def get_subg_info(sub_g):
    global DPU_NAME
    global SUBGRAPH_DB

    name = sub_g.get_name().strip()
    idx_w = subg_to_id(sub_g, "workload")

    workload = 0
    op_num = 0
    load_io_size = 0
    save_io_size = 0
    i_tensor_shape = []
    o_tensor_shape = []
    device = "unknow"

    i_tensors = sub_g.get_input_tensors()
    for it in i_tensors:
        i_tensor_shape.append(it.dims)
    o_tensors = sub_g.get_output_tensors()
    for ot in o_tensors:
        o_tensor_shape.append(ot.dims)

    if (sub_g.has_attr("dpu_name")):
        DPU_NAME = sub_g.get_attr("dpu_name")

    op_num = sub_g.get_op_num()
    if (sub_g.has_attr("workload")):
        workload = sub_g.get_attr("workload")

    if (sub_g.has_attr("device")):
        device = sub_g.get_attr("device")

    if (sub_g.has_attr("ac_code")):
        ac = sub_g.get_attr("ac_code")
        load_incs = [i for i in ac if i.startswith('LOAD')]
        for i in load_incs:
            load_io_size += inc_get_io_tensor_size(i)
        save_incs = [i for i in ac if i.startswith('SAVE')]
        for i in save_incs:
            save_io_size += inc_get_io_tensor_size(i)

    elif (sub_g.has_attr("mc_code")):
        mc = sub_g.get_attr("mc_code")
        load_io_size, save_io_size = process_mc(mc)

    subg_info = {idx_w:
                 {"name": name,
                     "device": device,
                     "workload": workload,
                     "op_num": op_num,
                     "load_io_size": load_io_size,
                     "save_io_size": save_io_size,
                     "i_tensor_shape": i_tensor_shape,
                     "o_tensor_shape": o_tensor_shape}}
    if idx_w not in SUBGRAPH_DB.keys():
        SUBGRAPH_DB.update(subg_info)
    else:
        logging.warning("SubGraph index duplication: [%s]" % idx_w)

    for sub in sub_g.get_children():
        get_subg_info(sub)


def xmodel_get_info(xmodel: []):
    SUBGRAPH_DB.clear()

    for m in xmodel:
        if not os.path.exists(m):
            raise("Error XModel Path: %s" % m)

        graph = xir.Graph.deserialize(m)
        root = graph.get_root_subgraph()
        get_subg_info(root)

    return SUBGRAPH_DB




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import tracer.tracerBase


class cmdTracer(tracer.tracerBase.Tracer):
    def __init__(self):
        super().__init__('cmd', source=[], compatible={
            'machine': ["x86_64", "aarch64"]})
        self.cmd = []

    def prepare(self, option: dict, debug: bool):
        """
        { control: {
                cmd: [],
                xat: {}
            }
        }
        """
        "Handle Input Options"
        self.cmd = option.get('control', {})

        if len(self.cmd) == 0:
            assert()

        "Handle Output Options"
        return {}

    def process(self, data, t_range=[]):
        pass

    def getData(self):
        return self.cmd


tracer.tracerBase.register(cmdTracer())



#!/usr/bin/python3
# -*- coding:utf-8 -*-

# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import re
import json
import pickle
import tracer.tracerBase


class vartTracer(tracer.tracerBase.Tracer):
    def __init__(self):
        super().__init__('vart', source=[
            'tracepoint'], compatible={'machine': ["x86_64", "aarch64"]})
        self.traceData = []
        self.timesync = 0

    def process(self, data, t_range=[]):
        for s in self.source:
            d = data.get(s, None)
            if d is None:
                continue
            self.traceData = d

            self.timesync = 0

            for trace in d:
                if trace.get('classname', "") == "trace_timesync":
                    self.timesync = float(
                        trace.get('steady_clock', 0)) - float(trace.get('xrt_ts', 0))
                    break

    def start(self):
        super().start()

    def stop(self):
        super().stop()

    def prepare(self, options: dict, debug: bool):
        self.saveTo = None
        if debug:
            self.saveTo = "./%s.trace" % self.name

    def getData(self):
        if self.saveTo != None:
            with open(self.saveTo, "w+t") as save:
                for t in self.traceData:
                    save.write(str(t))
                    save.write('\n')
        return self.traceData


tracer.tracerBase.register(vartTracer())




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import vaitraceSetting
import tracer.tracerBase
from subprocess import Popen, PIPE
import os

symbolFmt = "%-27s%-35s%-45s%9s\n"
TRACE_NAME_MASK = "__cln2_"

vaiLibPath = ["/usr/lib/"]
vaiLibs = [
    "libvart-buffer-object.so",
    "libvart-dpu-controller.so",
    "libvart-dpu-runner.so",
    "libvart-runner.so",
    "libvart-util.so",
    "libvart-xrt-device-handle.so",

    "libvitis_ai_library-math.so",
    "libvitis_ai_library-general1.so",
    "libvitis_ai_library-dpu_task.so",

    "libvitis_ai_library-tfssd.so",
    "libvitis_ai_library-facerecog.so",
    "libvitis_ai_library-RGBDsegmentation.so",
    "libvitis_ai_library-reidtracker.so",
    "libvitis_ai_library-platerecog.so",
    "libvitis_ai_library-facedetectrecog.so",
    "libvitis_ai_library-pmg.so",
    "libvitis_ai_library-ultrafast.so",
    "libvitis_ai_library-covid19segmentation.so",
    "libvitis_ai_library-3Dsegmentation.so",
    "libvitis_ai_library-facedetect.so",
    "libvitis_ai_library-openpose.so",
    "libvitis_ai_library-facequality5pt.so",
    "libvitis_ai_library-pointpillars.so",
    "libvitis_ai_library-platenum.so",
    "libvitis_ai_library-ssd.so",
    "libvitis_ai_library-yolov2.so",
    "libvitis_ai_library-rcan.so",
    "libvitis_ai_library-solo.so",
    "libvitis_ai_library-platedetect.so",
    "libvitis_ai_library-pointpillars_nuscenes.so",
    "libvitis_ai_library-clocs.so",
    "libvitis_ai_library-pointpainting.so",
    "libvitis_ai_library-facefeature.so",
    "libvitis_ai_library-lanedetect.so",
    "libvitis_ai_library-arflow.so",
    "libvitis_ai_library-multitaskv3.so",
    "libvitis_ai_library-centerpoint.so",
    "libvitis_ai_library-carplaterecog.so",
    "libvitis_ai_library-facelandmark.so",
    "libvitis_ai_library-refinedet.so",
    "libvitis_ai_library-ofa_yolo.so",
    "libvitis_ai_library-ocr.so",
    "libvitis_ai_library-retinaface.so",
    "libvitis_ai_library-vehicleclassification.so",
    "libvitis_ai_library-yolovx.so",
    "libvitis_ai_library-posedetect.so",
    "libvitis_ai_library-medicalsegmentation.so",
    "libvitis_ai_library-bcc.so",
    "libvitis_ai_library-medicalsegcell.so",
    "libvitis_ai_library-hourglass.so",
    "libvitis_ai_library-reid.so",
    "libvitis_ai_library-multitask.so",
    "libvitis_ai_library-efficientdet_d2.so",
    "libvitis_ai_library-textmountain.so",
    "libvitis_ai_library-polypsegmentation.so",
    "libvitis_ai_library-mnistclassification.so",
    "libvitis_ai_library-c2d2_lite.so",
    "libvitis_ai_library-fusion_cnn.so",
    "libvitis_ai_library-medicaldetection.so",

    "libxir.so",
    "libxnnpp-xnnpp.so"
]

"""
traceFuncTable--|--traceFunc--|--prop: top
                |             |--func--|--symbol:  "_Z13run_resnet_50P8dpu_taskRKN2cv3MatENSt7__cxx1112basic_strngI"
                |             |--name--|--name:    "run_resnet_50"
                |                      |--offset:  0x448c
                |                      |--libpath: "/root/lttng/resnet50/test_dnndk_resnet50"
"""


class traceFunc:
    def __init__(self, name, sym, path, offset, section="T"):
        self.name = name
        self.symbol = sym
        self.libPath = path
        self.offset = offset
        self.section = section

    def __str__(self):
        return symbolFmt % \
            (self.name[:25].replace(TRACE_NAME_MASK, '::'),
             self.symbol[:33], self.libPath, hex(self.offset))


class traceSection:
    def __init__(self, _name):
        self.prop = _name
        self.functions = []

    def __str__(self):
        str = "Section: [%s]\n" % self.prop
        str += symbolFmt % ("Function", "SymbolName", "LibPath", "Offset")

        for fun in self.functions:
            str += fun.__str__()

        return str


class symbolTable:
    def __init__(self, _exe):
        self.imgStartFlag = ""
        self.imgEndFlag = ""
        self.exe = os.path.abspath(_exe)
        self.table = []
        self.libs = []
        self.libs.append(self.exe)

        if os.path.exists(self.exe) == False:
            raise RuntimeError("Executable file not exists: [%s]" % self.exe)

    def getSymbols(self, extLibs=""):
        print("Analyzing symbol tables...")

        for libPath in self.libs:

            print("%d / %d" %
                  (self.libs.index(libPath) + 1, len(self.libs)), end="\r")
            if (self.libs.index(libPath) + 1 == len(self.libs)):
                print("")

            res = Popen("nm " + libPath + ' | egrep ".+ [T|t] .+"',
                        shell=True, stderr=PIPE, stdout=PIPE).stdout.readlines()

            symbols = []

            if len(res) > 0:
                """Get symbol tables from nm"""
                for s in res:
                    if len(s.split()) == 3:
                        symbols.append(s.decode().strip('\n').split())
            else:
                """For stripped .so, try get symbols from objdump -T"""
                res = Popen("objdump -T " + libPath,
                            shell=True, stderr=PIPE, stdout=PIPE).stdout.readlines()

                for s in res:
                    if len(s.split()) == 7:
                        deS = s.decode().strip('\n').split()

                        if deS[3] != ".text":
                            continue

                        symbols.append([deS[0], "T", deS[6]])

            self.table.append({"path": libPath, "symbols": symbols})

    def getLibs(self, extExe=""):
        if extExe == "":
            lib = self.exe
        else:
            lib = extExe
            extLibs = []

        lddRes = Popen("ldd " + lib, shell=True, stderr=PIPE,
                       stdout=PIPE).stdout.readlines()

        for line in lddRes:
            line = line.decode().strip('\n').strip('\t')
            index = line.find(" => ")
            if index < 0:
                continue
            libName = line[0:index]
            libPath = line[index+4:]
            libPath = libPath[0:libPath.find(" (")]

            if extExe == "":
                self.libs.append(libPath)
            else:
                extLibs.append(libPath)

        for l1 in vaiLibs:
            match = False
            for l2 in self.libs:
                if l2.find(l1) > 0:
                    match = True
                    break
            if match == False:
                for builtInLibPath in vaiLibPath:
                    if os.path.exists(os.path.join(builtInLibPath, l1)):
                        self.libs.append(os.path.join(builtInLibPath, l1))

        if extExe != "":
            return extLibs

    def getSymbolTable(self):
        self.getLibs()
        self.getSymbols()

        return [lib for lib in self.table if len(lib["symbols"]) > 0]


class traceConf:
    def __init__(self, _traceList, _symbolTable=None, _debug=False):
        self.traceSections = []
        self.symbolTable = _symbolTable
        self.parsed = False
        self.debug = _debug
        self.traceList = _traceList
        self.ftraceSymMap = {}

    def __str__(self):
        if not self.parsed:
            return "un-parsed conf"

        str = "Conf %s: %d section(s)\n" % (self.conf,
                                            len(self.traceSections))
        str += "===========================================================\n"

        for s in self.traceSections:
            str += s.__str__()

        return str

    def matchSymbol(self, _funcName, _table=""):
        if _table == "":
            table = self.symbolTable
        else:
            table = _table

        symbolList = list()
        idx = 0
        searchInLib = ""

        if (_funcName.find("@") > 0):
            searchInLib = _funcName.split('@')[1]
            _funcName = _funcName.split('@')[0]

        for lib in table:
            if (searchInLib != ""):
                if (lib["path"].find(searchInLib) < 0):
                    continue

            for item in lib["symbols"]:
                symbol = item[2]

                """Match C++ function"""
                if symbol.startswith("_Z"):
                    """
                    Match naming space
                    cv::imread -> 2cv6imread
                    """
                    funcName = "".join(["%d%s" % (len(s), s)
                                        for s in _funcName.split('::')])
                else:
                    funcName = _funcName

                """Exact Maching..."""
                em = False
                """_func == _ZN5vitis2ai27classification_post_processERKSt6vectorINS0_7library11InputTensorESaIS3_EERKS1_INS2_12OutputTensorESaIS8_EERKNS0_5proto13DpuModelParamEm"""
                if _funcName.startswith("_Z"):
                    funcName = _funcName
                    em = True

                if (symbol.find(funcName) >= 0) or (em and (symbol == funcName)):
                    offset = int(item[0], 16)
                    sym = symbol
                    path = lib["path"]

                    if em:
                        try:
                            showName = os.popen(
                                "c++filt -p %s" % _funcName).readlines()[0].strip()
                        except:
                            showName = _funcName
                    else:
                        if idx > 0:
                            showName = _funcName + "_%d" % idx
                        else:
                            showName = _funcName

                    ftraceSymName = str(
                        "_fun_" + str(hash(showName) % 10000000))
                    self.ftraceSymMap.update({ftraceSymName: showName})
                    self.ftraceSymMap.update({showName: ftraceSymName})

                    idx += 1

                    """Origin name for show"""
                    #symbolList.append(traceFunc(_funcName.split("::")[-1], sym, path, offset))
                    symbolList.append(
                        traceFunc(ftraceSymName, sym, path, offset))

        return symbolList

    def parseTableJson(self):
        if self.parsed == True:
            return self.traceSections

        conf = self.traceList

        """Add and match functions to be traced get keys that starts with 'trace'"""
        for section in [s for s in conf.keys() if s.startswith('trace')]:
            if conf[section] is None:
                continue

            """Remove duplicates"""
            conf[section] = list(set(conf[section]))

            sec = traceSection(section)

            for func in conf[section]:
                print("%d / %d" %
                      (conf[section].index(func) + 1, len(conf[section])), end="\r")

                if conf[section].index(func) + 1 == len(conf[section]):
                    print("")

                match = self.matchSymbol(func, self.symbolTable)

                if len(match) > 0:
                    for s in match:
                        sec.functions.append(s)
                else:
                    if self.debug:
                        ask("Function [%s] not found in symbol table" % func)

            self.traceSections.append(sec)

        self.parsed = True
        return self.traceSections


class functionTracer(tracer.tracerBase.Tracer):
    def __init__(self):
        super().__init__('function', source=["ftrace"], compatible={
            'machine': ["x86_64", "aarch64"]})
        self.ftraceSymMap = {}

    def prepare(self, options: dict, debug: bool):
        """Process configure file & symbol table"""
        table = symbolTable(options.get(
            'control').get('cmd')[0]).getSymbolTable()
        optionFunction = options.get('tracer', {}).get('function', {})
        traceListOri = optionFunction.get('traceList')

        z = traceConf(traceListOri, table)

        traceList = []
        traceListOri = z.parseTableJson()
        self.ftraceSymMap = z.ftraceSymMap

        "Handle Input Options"

        for section in traceListOri:
            if not section.prop.startswith("trace_"):
                continue
            for func in section.functions:
                traceList.append(
                    [func.name, func.libPath, "0x%x" % func.offset])

        saveTo = None
        if debug:
            saveTo = './function.trace'

        "Handle Output Options"
        optForFunction = {
            "collector": {
                "ftrace": {
                    "function": {
                        "name": "function",
                        "type": "uprobe",
                        "saveTo": saveTo,
                        "traceList": traceList
                    }
                }
            }
        }

        return optForFunction

    def compatible(self, platform: {}):
        if super().compatible(platform) == False:
            return False

        """Do some tests"""
        return vaitraceSetting.checkFtrace()

    def process(self, data, t_range=[]):
        """"test_performanc-19925 [003] .... 110337.353720: _fun_9351538_entry: (0xffff8553b0d8)"""
        data = [l for l in data.get('ftrace', {}).get(
            self.name) if not l.startswith('#')]
        data_out = []
        for l in data:
            ftraceSym = l.strip().split()[4].rsplit('_', 1)[0]
            data_out.append(l.replace(ftraceSym, self.ftraceSymMap[ftraceSym]))
            # for k in self.ftraceSymMap.keys():
            #    if l.find(k) > 0:
            #        data_out.append(l.replace(k, self.ftraceSymMap[k]))
        self.data = data_out

    def getData(self):
        return self.data


tracer.tracerBase.register(functionTracer())



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import time
import signal
import time
import logging
from subprocess import Popen, PIPE

import collector
import tracer
import vaitraceCfgManager
import vaitraceSetting


force_exit = False


def handler(signum, frame):
    logging.info("Killing process...")
    logging.info("Processing trace data, please wait...")
    global force_exit
    if force_exit:
        logging.error("Force exit...")
        exit(-1)
    force_exit = True


def shell_find_exec_path(_exe):

    exe_abs_path = ""

    if os.path.exists(os.path.abspath(_exe)):
        """1. search in cur dir"""
        exe_abs_path = os.path.abspath(_exe)

    else:
        """2. search in path via 'which'"""
        which_cmd = ["which", _exe]
        p = Popen(which_cmd, stdout=PIPE, stderr=PIPE)
        res = p.stdout.readlines()
        if len(res) > 0:
            exe_abs_path = res[0].strip().decode()

    if os.path.exists(exe_abs_path) == False:
        raise RuntimeError("Executable file not exists [%s]" % _exe)

    logging.info("Executable file: %s" % exe_abs_path)

    return exe_abs_path


def run(globalOptions: dict):
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)
    options = globalOptions

    """Help find the path of cmd in PATH"""
    cmd = options.get('control').get('cmd')[0]
    options['control']['cmd'][0] = shell_find_exec_path(cmd)

    if options.get('cmdline_args').get('bypass', False):
        cmd = options.get('control').get('cmd')
        logging.info("Bypass vaitrace, just run cmd")
        proc = Popen(cmd)
        proc.wait()
        exit(0)

    """Preparing"""
    tracer.prepare(options)
    tracer.start()

    """requirememt format: ["tracerName", "tracerName1", "hwInfo", ...]"""
    collector.prepare(options, tracer.getSourceRequirement())
    collector.start()

    """Start Running"""
    cmd = options.get('control').get('cmd')
    timeout = options.get('control').get('timeout')
    proc = Popen(cmd)

    options['control']['launcher'] = "cpp"
    options['control']['pid'] = proc.pid
    options['control']['time'] = time.strftime(
        "%Y-%m-%d %H:%M:%S", time.localtime())

    if timeout <= 0:
        proc.wait()
    else:
        while timeout > 0:
            time.sleep(1)
            timeout -= 1
            p = proc.poll()
            if p is not None:
                break

    logging.info("vaitrace time out, stopping process...")
    proc.send_signal(signal.SIGINT)
    proc.wait()
    collector.stop()
    tracer.stop()

    tracer.process(collector.getData())



"""
Copyright 2021 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import datetime
import time
import numpy as np
import pandas as pd
import os
from os import environ, getenv

from tensorflow.python.keras.preprocessing.text import Tokenizer
from tensorflow.python.keras.preprocessing.sequence import pad_sequences
from tensorflow.python.keras import layers
from tensorflow.python.keras import Sequential
from tensorflow.python import keras
from sklearn.model_selection import train_test_split
import tensorflow as tf
import vart
import xir


current_dir = os.path.dirname(__file__)
filename = os.path.join(current_dir, "data", "car_rental_training_data.csv")

data = pd.read_csv(filename, sep=';')
data.head()

complain_data = data[['Customer_Service', 'Satisfaction']]

# Generate rnn graph
runners = []
batches = []

device_name = getenv("TARGET_DEVICE", default="").upper()
assert(device_name in ("U50LV", "U25")), "TARGET_DEVICE should be U50LV/U25"

num_cores = 1
if device_name == "U50LV":
    num_cores = 2
    models = [os.path.join(current_dir, "data", file)
              for file in ["compiled_batch_3.xmodel", "compiled_batch_4.xmodel"]]
elif device_name == "U25":
    num_cores = 1
    models = [os.path.join(current_dir, "data", file)
              for file in ["compiled_batch_1.xmodel"]]

num_sequences = 25
output_seq_dim = 100

# Setup the Runners
for i in range(num_cores):
    print("[INFO] Creating Runner on core:{}".format(i))
    graph = xir.Graph.deserialize(models[i])
    runners.append(vart.Runner.create_runner(
        graph.get_root_subgraph(), "run"))
    inputTensors = runners[i].get_input_tensors()
    outputTensors = runners[i].get_output_tensors()
    batch_size, num_frames, runner_in_seq_len = tuple(inputTensors[0].dims)
    _, _, runner_out_seq_len = tuple(outputTensors[0].dims)
    batches.append(batch_size)

out_pos = graph.get_root_subgraph().get_attr('output_fix2float')
in_pos = graph.get_root_subgraph().get_attr('input_float2fix')

max_features = 500

def quanti_convert_float_to_int16(data, fix_pos):
    amp = 2**fix_pos
    max = 2**(16-1)
    output = data * amp
    output = np.clip(output, -max, max - 1)
    output = np.floor(output)
    output = output.astype(np.int16)
    return output


def quanti_convert_int16_to_float(data, fix_pos):
    amp = 2**fix_pos
    output = data.astype(np.float32)
    output = data / amp
    return output


for idx, row in complain_data.iterrows():
    row[0] = row[0].replace('rt', ' ')

tokenizer = Tokenizer(num_words=max_features, split=' ')
tokenizer.fit_on_texts(complain_data['Customer_Service'].values)
X = tokenizer.texts_to_sequences(complain_data['Customer_Service'].values)

maxlen = 50
X = pad_sequences(X, maxlen=maxlen)

Y = complain_data['Satisfaction'].values
X_train, X_test, Y_train, Y_test = train_test_split(
    X, Y, test_size=0.33, random_state=42)

# print(X_train.shape,Y_train.shape)
# print(X_test.shape,Y_test.shape)

embedding_vector_length = 32

model = Sequential()
model.add(layers.Embedding(max_features,
          embedding_vector_length, input_length=maxlen))
model.add(layers.Conv1D(filters=32, kernel_size=3,
          padding='same', activation='relu'))
model.add(layers.MaxPooling1D(pool_size=2))
model.add(layers.LSTM(100, recurrent_activation='sigmoid'))
model.add(layers.Dense(1, activation='sigmoid'))
model.compile(loss='binary_crossentropy',
              optimizer='adam', metrics=['accuracy'])
model.summary()

filename = os.path.join(current_dir, 'data', 'complain_model.h5')
is_training = False
if is_training:
    model.fit(X_train, Y_train, validation_data=(
        X_test, Y_test), epochs=20, batch_size=64)

    # Evaluate the model
    scores = model.evaluate(X_test, Y_test, verbose=0)
    print("Evaluation Accuracy: %.2f%%" % (scores[1]*100))
    model.save(filename, save_format='tf')
else:
    model.load_weights(filename)

t1 = time.time()

lstm_upstream = tf.keras.Model(
    inputs=model.input, outputs=model.get_layer('max_pooling1d').output)
lstm_input = lstm_upstream.predict(X_test, batch_size=8)
# print(lstm_input.shape)

num_records = lstm_input.shape[0]
quantized_lstm_input = quanti_convert_float_to_int16(
    lstm_input.reshape(num_records * 25*32), in_pos).reshape((num_records, 25*32))
lstm_output = np.zeros((num_records, 25*100), dtype=np.int16)


runner_idx = count = 0

outputTensors = runners[0].get_output_tensors()
_, _, runner_out_seq_len = tuple(outputTensors[0].dims)

quantized_input = quantized_lstm_input.view()
out_np = lstm_output.view()

print("[INFO] Running execute_async..")

while count < len(quantized_input):
    batch_size = batches[runner_idx]
    input_data = quantized_input[count:count+batch_size]
    batch_size = input_data.shape[0]

    input_data = input_data.reshape(batch_size, num_sequences, runner_in_seq_len)
    output_data = np.empty((batch_size, num_sequences, runner_out_seq_len), dtype=np.int16)
    job_id = runners[runner_idx].execute_async([input_data], [output_data], True)
    runners[runner_idx].wait(job_id)
    out_np[count:count+batch_size, ...] = output_data[..., :output_seq_dim].reshape(
            batch_size, num_sequences*output_seq_dim)

    count += batch_size
    runner_idx = (runner_idx + 1) % num_cores

while runners:
    del runners[0]

lstm_output = quanti_convert_int16_to_float(lstm_output, out_pos)
lstm_output = lstm_output.reshape((num_records, 25, 100))[:, -1, :]

lstm_downstream = Sequential()
lstm_downstream.add(layers.Dense(1, activation='sigmoid'))
lstm_downstream.compile(loss='binary_crossentropy',
                        optimizer='adam', metrics=['accuracy'])
lstm_downstream.build((1, 100))
lstm_downstream.layers[0].set_weights(model.get_layer('dense').get_weights())
score = lstm_downstream.evaluate(lstm_output, Y_test, verbose=0)
t2 = time.time()
print('Accuracy:', score[1])
print('E2E Time:', t2-t1)




# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class statItem:
    def __init__(self, name, timeunit="ms"):
        self.name = name
        self.runs = 0
        self.min_t = float("inf")
        self.max_t = 0.0
        self.ave_t = 0.0
        self.total_t = 0.0
        self.timeunit = timeunit

    def add(self, time):
        self.runs += 1
        self.total_t += time
        if time < self.min_t:
            self.min_t = time
        if time > self.max_t:
            self.max_t = time
        self.ave_t = self.total_t / self.runs

    def setTimeUnit(self, _timeunit):
        if _timeunit == "s" or _timeunit == "ms" or _timeunit == "us":
            self.timeunit = _timeunit

    def __str__(self):
        if self.timeunit == "ms":
            tu = 1000
        elif self.timeunit == "us":
            tu = 1000 * 1000
        else:
            tu = 1

        min_t = self.min_t * tu
        max_t = self.max_t * tu
        ave_t = self.ave_t * tu
        total_t = self.total_t * tu

        return "%s,%d,CPU,%.3f,%.3f,%.3f,\n" % (self.name, self.runs, min_t, ave_t, max_t)


class statTable:
    def __init__(self, name):
        self.items = {}
        self.name = name

    def add(self, name, time):
        if name not in self.keys():
            self.items.update({name: statItem(name)})
        self.items.get(name).add(time)

    def keys(self):
        return self.items.keys()

    def output(self, fmt="csv", timeunit="ms"):
        csv = []
        #csv.append(self.name + " " + "Summary\n")
        #csv.append("Function Name,Number Of Runs,Minimum Time (ms),Maximum Time (ms),Average Time (ms),\n")
        for k in self.items.keys():
            self.items[k].setTimeUnit(timeunit)
            csv.append(str(self.items[k]))
        return csv


"""
DPU Summary
Kernel Name,Number Of Runs,CU Full Name,Minimum Time (ms),Maximum Time (ms),Average Time (ms),
"""



"""
Copyright 2019 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List
import cv2
import numpy as np
import sys
import math
import xir
import vart

"""
 pre-process
"""
_B_MEAN = 104.0
_G_MEAN = 107.0
_R_MEAN = 123.0
MEANS = [_B_MEAN,_G_MEAN,_R_MEAN]
SCALES = [1.0, 1.0, 1.0]

def preprocess_one_image(image_path, width, height, means , scales, fixpos):
   image = cv2.imread(image_path)
   image = cv2.resize(image,(width, height))
   B, G, R = cv2.split(image)
   fix_scale = 2 ** fixpos
   B = (B - means[0]) * (scales[0] * fix_scale)
   G = (G - means[1]) * (scales[1] * fix_scale)
   R = (R - means[2]) * (scales[2] * fix_scale)
   image = cv2.merge([B, G, R])
   image = image.astype(np.int8)
   return image

"""
 post-process
"""
def CPUCalcSoftmax(data, size):
    sum = 0.0
    result = [0 for i in range(size)]
    for i in range(size):
        result[i] = math.exp(data[i])
        sum += result[i]
    for i in range(size):
        result[i] /= sum
    return result

def TopK(datain, size, filePath):
    cnt = [i for i in range(size)]
    pair = zip(datain, cnt)
    pair = sorted(pair, reverse=True)
    softmax_new, cnt_new = zip(*pair)
    fp = open(filePath, "r")
    data1 = fp.readlines()
    fp.close()
    for i in range(5):
        idx = 0
        for line in data1:
            if idx == cnt_new[i]:
                print("Top[%d] %d %s" % (i, idx, (line.strip)("\n")))
            idx = idx + 1

"""
 obtain dpu subgrah
"""
def get_child_subgraph_dpu(graph: "Graph") -> List["Subgraph"]:
    assert graph is not None, "'graph' should not be None."
    root_subgraph = graph.get_root_subgraph()
    assert (
        root_subgraph is not None
    ), "Failed to get root subgraph of input Graph object."
    if root_subgraph.is_leaf:
        return []
    child_subgraphs = root_subgraph.toposort_child_subgraph()
    assert child_subgraphs is not None and len(child_subgraphs) > 0
    return [
        cs
        for cs in child_subgraphs
        if cs.has_attr("device") and cs.get_attr("device").upper() == "DPU"
    ]


def main(argv):
    g = xir.Graph.deserialize(argv[1])
    subgraphs = get_child_subgraph_dpu(g)
    assert len(subgraphs) == 1  # only one DPU kernel
    runner = vart.RunnerExt.create_runner (subgraphs[0], "run")
    input_tensor_buffers = runner.get_inputs()
    output_tensor_buffers = runner.get_outputs()

    input_ndim = tuple(input_tensor_buffers[0].get_tensor().dims)
    batch = input_ndim[0]
    width = input_ndim[1]
    height = input_ndim[2]
    fixpos = input_tensor_buffers[0].get_tensor().get_attr("fix_point")
    image = preprocess_one_image(argv[2], width, height, MEANS, SCALES, fixpos)
    input_data = np.asarray(input_tensor_buffers[0])
    input_data[0] = image

    job_id = runner.execute_async(input_tensor_buffers, output_tensor_buffers)
    runner.wait(job_id)

    pre_output_size = int(output_tensor_buffers[0].get_tensor().get_data_size() / batch)
    output_data = np.asarray(output_tensor_buffers[0])
    sfm_data = CPUCalcSoftmax(output_data[0], pre_output_size)
    #print(sfm_data)
    TopK(sfm_data, pre_output_size, "./words.txt")
    del runner


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("please  model file and input file.")
    else:
        main(sys.argv)



#!/usr/bin/python

# Copyright 2019 Xilinx Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import glob
import argparse
import statistics as stat

KEYS = ["INSTR_COMPUTE", "INSTR_UPLOAD", "REG_COMPUTE", "INPUT_UPLOAD", "XRT_RUN", "OUTPUT_DOWNLOAD", "EXECUTE_ASYNC"]

class ProfData:
    def __init__(self, device, model, batch, lats):
        self.device = device
        self.model = model
        self.batch = batch
        self.lats = lats

    def mean(self, key):
        return stat.mean(self.lats[key])

    def stddev(self, key):
        return stat.pstdev(self.lats[key])

def parse_file_name(logfile):
    ''' Parse a filename of the form sth_device_model_batch.lat
        and return (device, model, batch)
    '''
    fname = os.path.basename(logfile)
    f = os.path.splitext(fname)[0]
    tokens = f.split('_')
    if len(tokens) > 3:
        return (tokens[-3], tokens[-2], tokens[-1])
    else:
        return ("", "", "")

def parse_file(logfile):
    ''' Parse a file to collect item-specific latencies '''
    assert(os.path.exists(logfile))
    lats = dict((key, []) for key in KEYS)
    with open(logfile, "r") as f:
        for line in f:
            for key, _ in lats.items():
                if key in line:
                    lat_str = line.split()[-1].rstrip("us")
                    lat = float(lat_str)
                    lats[key].append(lat)
                    break
    return lats

def generate_prof_data(device, model, batch, logfile):
    lats = parse_file(logfile)
    return ProfData(device, model, batch, lats)

def generate_prof_datas(logdir, ext="lat"):
    ifiles = glob.glob(logdir + "/*." + ext)
    profs = []
    for f in ifiles:
        print(f)
        device, model, batch = parse_file_name(f)
        prof = generate_prof_data(device, model, batch, f)
        profs.append(prof)
    return profs

def save_as_csv(pdatas, outfile):
    row_headers = ["Model", "Device", "Batch"] + KEYS
    nrows = len(row_headers)
    data = [[item] for item in row_headers]
    for pdata in pdatas:
        data[0].append(pdata.model)
        data[1].append(pdata.device)
        data[2].append(pdata.batch)
        for i, key in enumerate(KEYS, 3):
            mean = round(pdata.mean(key))
            stddev = round(pdata.stddev(key))
            data[i].append(u"{} \u00b1 {}".format(mean, stddev))
    with open(outfile, 'w') as f:
        for d in data:
            f.write(','.join(d) + '\n')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--log_dir", help="Directory containing all log files")
    parser.add_argument("--output_file", help="Path of csv file to save the result")
    args = parser.parse_args()
    pdatas = generate_prof_datas(args.log_dir)
    save_as_csv(pdatas, args.output_file)






# Copyright 2019 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from writer.parser.timelineUtil import *

hwCUPatt = re.compile(
    r"""
        .+
        cu_idx=(?P<cu_idx>\d+)
        """,
    re.X | re.M
)
"""for arm, trace_clock = boot"""
pattFtrace = re.compile(
    r"""
        (?P<taskComm>.+)\-
        (?P<taskPid>\d{1,})\s+\[
        (?P<cpuID>\d{3})\].{4,}\s
        (?P<timeStamp>\d+\.\d{6})\:\s+
        (?P<func>[\w:]+)\:\s+
        (?P<info>.*)
        """,
    re.X | re.M
)


class ftraceEvent:
    def __init__(self, taskComm, taskPid, cpuId, timeStamp, func, info):
        self.taskComm = taskComm.strip()
        self.taskPid = int(taskPid)
        self.cpuId = int(cpuId)
        self.timeStamp = float(timeStamp)
        self.func = func.strip()
        self.info = info.strip()
        self.infoDetail = dict()
        self.isTarget = False
        self.dir = None

    def __str__(self):
        return "%.6f:%17s-%4d@[%02d]:%18s: %s" %\
               (self.timeStamp, self.taskComm, self.taskPid,
                self.cpuId, self.func, self.info)

    def toTimelineEvent(self):
        if self.func.endswith("_entry"):
            et = "start"
        elif self.func.endswith("_exit"):
            et = "done"
        else:
            et = "marker"

        ct = "CPU"
        cid = self.cpuId

        func = self.func.replace("_entry", "").replace("_exit", "")
        return vaiTimelineEvent(self.timeStamp, self.taskPid, et, ct, cid, func)


def ftraceParse(l, options):
    patt = pattFtrace

    """Selet trace clock"""
    tc = options.get('traceClock', None)
    if tc == 'x86-tsc':
        patt = pattFtraceTSC
    tmp = re.match(patt, l.strip())

    """Not matched"""
    if tmp is None:
        return None

    tmp = tmp.groups()
    return(ftraceEvent(*tmp))



#Copyright 2021 Xilinx Inc.
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

#MIT License
#
#Copyright (c) 2016 Gabriel Stanovsky
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

""" Usage:
    <file-name> --in=INPUT_FILE --batch-size=BATCH-SIZE --out=OUTPUT_FILE --model-path=MODEL_PATH [--cuda-device=CUDA_DEVICE] [--debug]
"""
# External imports
import os
import logging
from pprint import pprint
from pprint import pformat
from docopt import docopt
import json
import pdb
from tqdm import tqdm
import torch
from allennlp.common.params import Params
from allennlp.models.model import Model
from allennlp.models.archival import Archive, CONFIG_NAME, _WEIGHTS_NAME
from allennlp.predictors import Predictor
from allennlp.nn import util
from allennlp.data import Instance, Vocabulary
from collections import defaultdict
from operator import itemgetter
import functools
import operator
import time

# Local imports
from format_oie import format_extractions, Mock_token
#=-----

def chunks(l, n):
    """
    Yield successive n-sized chunks from l.
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]

def create_instances(model, sent):
    """
    Convert a sentence into a list of instances.
    """
    sent_tokens = model._tokenizer.tokenize(sent)
    # Find all verbs in the input sentence
    pred_ids = [i for (i, t) in enumerate(sent_tokens)
                if t.pos_ == "VERB"]

    # Create instances
    instances = [{"sentence": sent_tokens,
                  "predicate_index": pred_id}
                 for pred_id in pred_ids]

    return instances

def get_confidence(model, tag_per_token, class_probs):
    """
    Get the confidence of a given model in a token list, using the class probabilities
    associated with this prediction.
    """
    #tag_per_token = list([tag for tag in tag_per_token if tag != 'O'])
    token_indexes = [model._model.vocab.get_token_index(tag, namespace = "labels") for tag in tag_per_token]
    # Get probability per tag
    #probs = [class_prob[token_index] for token_index, class_prob in zip(token_indexes, class_probs) if token_index != 0]
    probs = [class_prob[token_index] for token_index, class_prob in zip(token_indexes, class_probs)]

    # Combine (product)
    if len(probs) == 0:
        prod_prob = 0.0
        print(tag_per_token)
    else:
        prod_prob = functools.reduce(operator.mul, probs)

    return prod_prob

def remove_pretrained_embedding_params(params: Params):
    if isinstance(params, Params):  # The model could possible be a string, for example.
        keys = params.keys()
        if "pretrained_file" in keys:
            del params["pretrained_file"]
        for value in params.values():
            if isinstance(value, Params):
                remove_pretrained_embedding_params(value)

def openie_model(serialization_dir, weights_file=None, cuda_device=-1):
    """
    Instantiates an already-trained model, based on the experiment
    configuration and some optional overrides.
    """
    # Load config
    config = Params.from_file(os.path.join(serialization_dir, CONFIG_NAME), "")
    config.loading_from_archive = True
    config = config.duplicate()

    weights_file = weights_file or os.path.join(serialization_dir, _WEIGHTS_NAME)

    # Load vocabulary from file
    vocab_dir = os.path.join(serialization_dir, "vocabulary")
    # If the config specifies a vocabulary subclass, we need to use it.
    vocab_params = config.get("vocabulary", Params({}))
    vocab_choice = vocab_params.pop_choice("type", Vocabulary.list_available(), True)
    vocab = Vocabulary.by_name(vocab_choice).from_files(vocab_dir)

    model_params = config.get("model")

    # The experiment config tells us how to _train_ a model, including where to get pre-trained
    # embeddings from.  We're now _loading_ the model, so those embeddings will already be
    # stored in our weights.  We don't need any pretrained weight file anymore, and we don't
    # want the code to look for it, so we remove it from the parameters here.
    remove_pretrained_embedding_params(model_params)
    model = Model.from_params(vocab=vocab, params=model_params)

    # If vocab+embedding extension was done, the model initialized from from_params
    # and one defined by state dict in weights_file might not have same embedding shapes.
    # Eg. when model embedder module was transferred along with vocab extension, the
    # initialized embedding weight shape would be smaller than one in the state_dict.
    # So calling model embedding extension is required before load_state_dict.
    # If vocab and model embeddings are in sync, following would be just a no-op.
    model.extend_embedder_vocab()

    model_state = torch.load(weights_file, map_location=util.device_mapping(cuda_device))

    model.load_state_dict(model_state)

    # Force model to cpu or gpu, as appropriate, to make sure that the embeddings are
    # in sync with the weights
    if cuda_device >= 0:
        model.cuda(cuda_device)
    else:
        model.cpu()

    return model, config

if __name__ == "__main__":
    # Parse command line arguments
    args = docopt(__doc__)
    inp_fn = args["--in"]
    batch_size = int(args["--batch-size"])
    out_fn = args["--out"]
    model_path = args["--model-path"]
    cuda_device = int(args["--cuda-device"]) if (args["--cuda-device"] is not None) \
                  else -1
    debug = args["--debug"]
    if debug:
        logging.basicConfig(level = logging.DEBUG)
    else:
        logging.basicConfig(level = logging.INFO)

    # Init OIE
    raw_model, config = openie_model(model_path)
    # insert nndct code
    archive = Archive(model=raw_model, config=config)
    model = Predictor.from_archive(archive, "open-information-extraction")
    # model = open_information_extraction_stanovsky_2018()

    # Move model to gpu, if requested
    if cuda_device >= 0:
        model._model.cuda(cuda_device)

    lines = [line.strip()
             for line in open(inp_fn, encoding = "utf8")]

    # process sentences
    logging.info("Processing sentences")
    oie_lines = []
    t1 = time.perf_counter()
    for chunk in tqdm(chunks(lines, batch_size)):
        oie_inputs = []
        for sent in chunk:
            oie_inputs.extend(create_instances(model, sent))
        if not oie_inputs:
            # No predicates in this sentence
            continue

        # Run oie on sents
        sent_preds = model.predict_batch_json(oie_inputs)

        # Collect outputs in batches
        predictions_by_sent = defaultdict(list)
        for outputs in sent_preds:
            sent_tokens = outputs["words"]
            tags = outputs["tags"]
            sent_str = " ".join(sent_tokens)
            assert(len(sent_tokens) == len(tags))
            predictions_by_sent[sent_str].append((outputs["tags"], outputs["class_probabilities"]))

        # Create extractions by sentence
        for sent_tokens, predictions_for_sent in predictions_by_sent.items():
            raw_tags = list(map(itemgetter(0), predictions_for_sent))
            class_probs = list(map(itemgetter(1), predictions_for_sent))

            # Compute confidence per extraction
            confs = [get_confidence(model, tag_per_token, class_prob)
                     for tag_per_token, class_prob in zip(raw_tags, class_probs)]

            extractions, tags = format_extractions([Mock_token(tok) for tok in sent_tokens.split(" ")], raw_tags)
            oie_lines.extend([extraction + f"\t{conf}" for extraction, conf in zip(extractions, confs)])
    t2 = time.perf_counter()
    print("E2E time: ", t2-t1)

    # Write to file
    logging.info(f"Writing output to {out_fn}")
    with open(out_fn, "w", encoding = "utf8") as fout:
        fout.write("\n".join(oie_lines))
    logging.info("DONE")



"""
Copyright 2019 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ctypes import *
from typing import List
import cv2
import numpy as np
import xir
import vart
import os
import math
import threading
import time
import sys

"""
Calculate softmax
data: data to be calculated
size: data size
return: softamx result
"""


def CPUCalcSoftmax(data, size, scale):
    sum = 0.0
    result = [0 for i in range(size)]
    for i in range(size):
        result[i] = math.exp(data[i] * scale)
        sum += result[i]
    for i in range(size):
        result[i] /= sum
    return result


def get_script_directory():
    path = os.getcwd()
    return path


"""
Get topk results according to its probability
datain: data result of softmax
filePath: filePath in witch that records the infotmation of kinds
"""


def TopK(datain, size, filePath):

    cnt = [i for i in range(size)]
    pair = zip(datain, cnt)
    pair = sorted(pair, reverse=True)
    softmax_new, cnt_new = zip(*pair)
    fp = open(filePath, "r")
    data1 = fp.readlines()
    fp.close()
    for i in range(5):
        idx = 0
        for line in data1:
            if idx == cnt_new[i]:
                print("Top[%d] %d %s" % (i, idx, (line.strip)("\n")))
            idx = idx + 1


"""
pre-process for resnet50 (caffe)
"""
_B_MEAN = 104.0
_G_MEAN = 107.0
_R_MEAN = 123.0
MEANS = [_B_MEAN, _G_MEAN, _R_MEAN]
SCALES = [1.0, 1.0, 1.0]


def preprocess_one_image_fn(image_path, fix_scale, width=224, height=224):
    means = MEANS
    scales = SCALES
    image = cv2.imread(image_path)
    image = cv2.resize(image, (width, height))
    B, G, R = cv2.split(image)
    B = (B - means[0]) * scales[0] * fix_scale
    G = (G - means[1]) * scales[1] * fix_scale
    R = (R - means[2]) * scales[2] * fix_scale
    image = cv2.merge([B, G, R])
    image = image.astype(np.int8)
    return image


SCRIPT_DIR = get_script_directory()
calib_image_dir = SCRIPT_DIR + "/../images/"
global threadnum
threadnum = 0
"""
run resnt50 with batch
runner: dpu runner
img: imagelist to be run
cnt: threadnum
"""


def runResnet50(runner: "Runner", img, cnt):
    """get tensor"""
    inputTensors = runner.get_input_tensors()
    outputTensors = runner.get_output_tensors()
    input_ndim = tuple(inputTensors[0].dims)
    pre_output_size = int(outputTensors[0].get_data_size() / input_ndim[0])

    output_ndim = tuple(outputTensors[0].dims)
    output_fixpos = outputTensors[0].get_attr("fix_point")
    output_scale = 1 / (2**output_fixpos)
    n_of_images = len(img)
    count = 0
    while count < cnt:
        runSize = input_ndim[0]
        """prepare batch input/output """
        inputData = [np.empty(input_ndim, dtype=np.int8, order="C")]
        outputData = [np.empty(output_ndim, dtype=np.int8, order="C")]
        """init input image to input buffer """
        for j in range(runSize):
            imageRun = inputData[0]
            imageRun[j, ...] = img[(count + j) % n_of_images].reshape(input_ndim[1:])
        """run with batch """
        job_id = runner.execute_async(inputData, outputData)
        runner.wait(job_id)
        """softmax&TopK calculate with batch """
        """Benchmark DPU FPS performance over Vitis AI APIs execute_async() and wait() """
        """Uncomment the following code snippet to include softmax calculation for model’s end-to-end FPS evaluation """
        #for j in range(runSize):
        #    softmax = CPUCalcSoftmax(outputData[0][j], pre_output_size, output_scale)
        #    TopK(softmax, pre_output_size, "./words.txt")

        count = count + runSize


"""
 obtain dpu subgrah
"""


def get_child_subgraph_dpu(graph: "Graph") -> List["Subgraph"]:
    assert graph is not None, "'graph' should not be None."
    root_subgraph = graph.get_root_subgraph()
    assert (root_subgraph
            is not None), "Failed to get root subgraph of input Graph object."
    if root_subgraph.is_leaf:
        return []
    child_subgraphs = root_subgraph.toposort_child_subgraph()
    assert child_subgraphs is not None and len(child_subgraphs) > 0
    return [
        cs for cs in child_subgraphs
        if cs.has_attr("device") and cs.get_attr("device").upper() == "DPU"
    ]


def main(argv):
    global threadnum

    listimage = os.listdir(calib_image_dir)
    threadAll = []
    threadnum = int(argv[1])
    i = 0
    global runTotall
    runTotall = len(listimage)
    g = xir.Graph.deserialize(argv[2])
    subgraphs = get_child_subgraph_dpu(g)
    assert len(subgraphs) == 1  # only one DPU kernel
    all_dpu_runners = []
    for i in range(int(threadnum)):
        all_dpu_runners.append(vart.Runner.create_runner(subgraphs[0], "run"))

    input_fixpos = all_dpu_runners[0].get_input_tensors()[0].get_attr("fix_point")
    input_scale = 2**input_fixpos
    """image list to be run """
    img = []
    for i in range(runTotall):
        path = os.path.join(calib_image_dir, listimage[i])
        img.append(preprocess_one_image_fn(path, input_scale))
    """
      The cnt variable is used to control the number of times a single-thread DPU runs.
      Users can modify the value according to actual needs. It is not recommended to use
      too small number when there are few input images, for example:
      1. If users can only provide very few images, e.g. only 1 image, they should set
         a relatively large number such as 360 to measure the average performance;
      2. If users provide a huge dataset, e.g. 50000 images in the directory, they can
         use the variable to control the test time, and no need to run the whole dataset.
    """
    cnt = 360
    """run with batch """
    time_start = time.time()
    for i in range(int(threadnum)):
        t1 = threading.Thread(target=runResnet50, args=(all_dpu_runners[i], img, cnt))
        threadAll.append(t1)
    for x in threadAll:
        x.start()
    for x in threadAll:
        x.join()

    del all_dpu_runners

    time_end = time.time()
    timetotal = time_end - time_start
    total_frames = cnt * int(threadnum)
    fps = float(total_frames / timetotal)
    print(
        "FPS=%.2f, total frames = %.2f , time=%.6f seconds"
        % (fps, total_frames, timetotal)
    )


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("usage : python3 resnet50.py <thread_number> <resnet50_xmodel_file>")
    else:
        main(sys.argv)
